# -*- coding: utf-8 -*-

import asyncio
import datetime
from ..observers import AnalyticsObserver
from ..functions.core_functions import *
from toloka.client.actions import ChangeOverlap
from toloka.client.collectors import AssignmentsAssessment
from toloka.client.conditions import AssessmentEvent
from toloka.util.async_utils import AsyncMultithreadWrapper
from toloka.streaming import AssignmentsObserver, Pipeline, PoolStatusObserver
from wasabi import Printer

# Set up Printer
msg = Printer(pretty=True, timestamp=True, hide_animation=True)


class TaskSequence:
    """
    This class allows defining a sequence of crowdsourcing tasks on Toloka.

    """
    def __init__(self, sequence, client):
        """
        This function initialises the TaskSequence class.

        Parameters:
            sequence: A list of objects that inherit from the CrowdsourcingTask class.
            client: A TolokaClient object with valid credentials.
        """
        # Set up attributes
        self.complete = False       # Tracks if all tasks have been completed
        self.sequence = sequence    # A list of CrowdsourcingTask objects
        self.client = client        # A Toloka Client object
        self.pipeline = None        # Placeholder for a Toloka Pipeline object

        msg.info(f'Creating a task sequence')

        # Verify that connections between tasks have been specified in the configuration
        verify_connections(self.sequence)

        msg.info(f'Printing tasks, inputs and outputs')
        create_pool_table(self.sequence)

        # Create the pipeline
        self.create_pipeline()

    def start(self):

        msg.info(f'Starting the task sequence')

        # Open all pools in the sequence that contain tasks. Note that pools without tasks cannot
        # be opened: they will be opened when tasks are added to them by these tasks.
        for task in self.sequence:

            if hasattr(task, 'tasks') and task.tasks is not None:

                self.client.open_pool(pool_id=task.pool.id)

            if hasattr(task, 'training') and task.training is not None:

                self.client.open_pool(pool_id=task.training.id)

        # Set up MetricCollectors for all pools
        metrics_collector = create_metrics(task_sequence=self)

        # Define an asynchronous function to run the MetricCollector and
        # the Pipeline objects at the same time
        async def run_sequence(metrics):

            # The 'ensure_future' method allows running the MetricCollector
            # in a fire-and-forget manner
            if metrics is not None:

                asyncio.ensure_future(metrics.run())

            # The Pipeline object needs to be awaited
            await asyncio.gather(self.pipeline.run())

        # Call the asynchronous function to start the collectors and pipeline
        asyncio.run(run_sequence(metrics=metrics_collector))

        # Collect pool statuses here
        status = []

        # Close all training and exam pools that may remain open after the pipeline finishes
        while not self.complete:

            # Loop over tasks in the sequence and filter out actions by checking for pools
            for task in (t for t in self.sequence if hasattr(t, 'pool')):

                # Get the last reason for closing a pool: should be 'MANUAL' or 'COMPLETED',
                # can also be None.
                pool_status = self.client.get_pool(pool_id=task.pool.id).last_close_reason

                if pool_status is not None and pool_status.value in ['COMPLETED', 'MANUAL']:

                    if self.client.get_pool(pool_id=task.pool.id).is_closed():

                        status.append(True)

                    else:

                        self.client.close_pool(pool_id=task.pool.id)

                        msg.info(f'Closed pool with ID {task.pool.id}')

                # Check if there is a training pool that should be closed
                if hasattr(task, 'training') and task.training is not None:

                    if self.client.get_pool(pool_id=task.training.id).is_closed():

                        status.append(True)

                    else:

                        self.client.close_pool(pool_id=task.training.id)

                        msg.info(f'Closed pool with ID {task.training.id}')

            for action in (a for a in self.sequence if hasattr(a, 'aggregator')):

                if action.complete == True:
                    
                    status.append(True)

            if all(status):

                # Wait for a minute to ensure that no new tasks are added to pools in the pipeline
                # before ending the task sequence
                time.sleep(60)

                if all(status):

                    self.complete = True

                    msg.good(f'Successfully completed the task sequence')

        # Check the outputs
        if self.complete:

            exit()

            # Check if tasks are supposed to output the results
            for task in self.sequence:

                if hasattr(task, 'pool'):

                    # Get the output DataFrame for each task; assign under 'output_data'
                    task.output_data = self.client.get_assignments_df(pool_id=task.pool.id)

                    # Check if the output should be written to disk
                    try:

                        if task.conf['actions'] is not None and 'output' in task.conf['actions']:

                            # Write the DataFrame to disk
                            task.output_data.to_csv(f'{task.name}_{task.pool.id}.csv')

                            msg.good(f'Wrote data for task {task.name} ({task.pool.id}) to disk.')

                    except KeyError:

                        pass

    def create_pipeline(self):

        # Create an asyncronous client
        async_client = AsyncMultithreadWrapper(self.client)

        # Loop over the tasks and create an AssignmentsObserver object for each task. Exam tasks
        # do not require AssignmentsObservers, because they do not create further tasks to be
        # forwarded.
        a_observers = {task.name: AssignmentsObserver(async_client, task.pool.id)
                       for task in self.sequence if hasattr(task, 'pool')
                       and not task.exam if hasattr(task, 'pool')}

        # Set up pool observers for monitoring all pool states, including exams. These observers
        # are needed for the Pipeline to run correctly: a Pipeline cannot run without observers.
        p_observers = {task.name: PoolStatusObserver(async_client, task.pool.id)
                       for task in self.sequence if hasattr(task, 'pool')}

        # Set up pool analytics observers for monitoring exam pools
        pa_observers = {task.name: AnalyticsObserver(async_client, task.pool,
                                                     max_performers=task.pool_conf['exam']['max_performers'])
                        for task in self.sequence if hasattr(task, 'pool')
                        and task.exam if hasattr(task, 'pool')}

        # Create a Toloka Pipeline object and register observers; call observers every 15 seconds
        self.pipeline = Pipeline(period=datetime.timedelta(seconds=15))

        # Create a dictionary of CrowdsourcingTasks in the pipeline keyed by their names
        tasks = {task.name: task for task in self.sequence}

        # Register each assignment observer with the Pipeline object
        for name, a_observer in a_observers.items():

            self.pipeline.register(observer=a_observer)
            
            msg.info(f'Registered an assignments observer for task {name} ({tasks[name].pool.id})')

        # Register each pool observer and actions
        for name, p_observer in p_observers.items():

            p_observer.on_closed(lambda pool: msg.info(f'Closed pool with ID {pool.id}'))
            p_observer.on_open(lambda pool: msg.info(f'Opened pool with ID {pool.id}'))

            current_task = tasks[name]

            if current_task.action_conf is not None:

                # If pool's input data comes from a source action (such as SeparateBBoxes), intialize tasks
                if 'data_source' in current_task.action_conf:

                    source = tasks[current_task.action_conf['data_source']]
                    source()

                # Register possible Aggregate-actions to activate when pool is closed
                if 'on_closed' in current_task.action_conf:

                    p_observer.on_closed(tasks[current_task.action_conf['on_closed']])

                    msg.info(f'Results from {name} will be aggregated with {tasks[current_task.action_conf["on_closed"]].name}')

            self.pipeline.register(observer=p_observer)

            msg.info(f'Registered a pool status observer for task {name} ({tasks[name].pool.id})')

        # Register each pool analytics observer
        for name, pa_observer in pa_observers.items():

            self.pipeline.register(observer=pa_observer)

            msg.info(f'Registered a pool analytics observer for task {name} ({tasks[name].pool.id})')

        # Loop over the assignment observers and get the actions configuration to determine task flow
        for name, observer in a_observers.items():

            # Get the CrowdsourcingTask object from the TaskSequence by matching its name
            current_task = tasks[name]

            # Check if actions have been configured
            if current_task.action_conf is not None:

                if 'on_accepted' in current_task.action_conf:

                    # Attempt to register the action with the AssignmentObserver. If a task is
                    # accepted, it will be sent to the CrowdsourcingTask object defined in the
                    # configuration.
                    try:

                        observer.on_accepted(tasks[current_task.action_conf['on_accepted']])

                        msg.info(f'Setting up a connection from {name} '
                                 f'to {tasks[current_task.action_conf["on_accepted"]].name} '
                                 f'on acceptance')

                    except KeyError:

                        raise_error(f'Could not find a CrowdsourcingTask object named '
                                    f'{current_task.action_conf["on_accepted"]} in the '
                                    f'TaskSequence. Please check the configuration '
                                    f'under the key "actions"!')

                if 'on_submitted' in current_task.action_conf:

                    try:

                        # Register the action with the AssignmentObserver. If a task is submitted,
                        # it will be sent to the CrowdsourcingTask object defined in the configuration.
                        observer.on_submitted(tasks[current_task.action_conf['on_submitted']])

                        msg.info(f'Setting up a connection from {name} '
                                 f'to {tasks[current_task.action_conf["on_submitted"]].name} '
                                 f'on submission')

                    except KeyError:

                        raise_error(f'Could not find a CrowdsourcingTask object named '
                                    f'{current_task.action_conf["on_submitted"]} in the '
                                    f'TaskSequence. Please check the configuration '
                                    f'under the key "actions"!')

                if 'on_rejected' in current_task.action_conf:

                    # Check if rejected assignments tasks should be routed to the same pool
                    if current_task.action_conf['on_rejected'] == name:

                        # Add a ChangeOverlap action to the pool, which returns this assignment into annotation queue.
                        current_task.pool.quality_control.add_action(
                            collector=AssignmentsAssessment(),
                            conditions=[AssessmentEvent == AssessmentEvent.REJECT],
                            action=(ChangeOverlap(delta=1, open_pool=True)))

                        # Update the pool configuration for the action to take place.
                        self.client.update_pool(current_task.pool.id, current_task.pool)

                        msg.info(f'Rejected tasks from pool {name} will be re-added to the pool')

                    else:

                        try:

                            # Register the action with the AssignmentObserver. If a task is rejected,
                            # it will be sent to the CrowdsourcingTask object defined in the configuration.
                            observer.on_rejected(tasks[current_task.action_conf['on_rejected']])

                            msg.info(f'Setting up a connection from {name}'
                                     f'to {tasks[current_task.action_conf["on_rejected"]].name} '
                                     f'on rejected')

                        except KeyError:

                            raise_error(f'Could not find a CrowdsourcingTask object named '
                                        f'{current_task.action_conf["on_rejected"]} in the '
                                        f'TaskSequence. Please check the configuration '
                                        f'under the key "actions"!')

                if ('on_result' in current_task.action_conf) and ('on_closed' not in current_task.action_conf):

                    # If current pool is not an action, use setup from current pool
                    if hasattr(current_task, 'pool'):

                        setup = current_task.conf['pool']['setup']

                    # If current pool is an action, use setup from source pool 
                    else:

                        setup = tasks[current_task.source].pool.setup

                    try:

                        # If tasks are not automatically accepted, forward using 'on_submitted'
                        if setup['auto_accept_solutions'] == False:

                            # Check if multiple forward pools are configured
                            if type(tasks[current_task.action_conf['on_result']]) == dict:

                                for pool in tasks[current_task.action_conf['on_result']].values():
                                    observer.on_submitted(pool)
                            
                            else:
                                observer.on_submitted(tasks[current_task.action_conf['on_result']])

                        # If tasks are automatically accepted, forward using 'on_accepted'
                        if setup['auto_accept_solutions'] == True:

                            # Check if multiple forward pools are configured
                            if type(tasks[current_task.action_conf['on_result']]) == dict:

                                for pool in tasks[current_task.action_conf['on_result']].values():
                                    observer.on_accepted(pool)

                            else:
                                
                                observer.on_accepted(tasks[current_task.action_conf['on_result']])

                        msg.info(f'Tasks from {name} will be forwarded with {tasks[current_task.action_conf["on_result"]].name} '
                                 f'on result according to configuration')

                    except KeyError:

                        raise_error(f'Could not find a CrowdsourcingTask object named '
                                    f'{current_task.action_conf["on_result"]} in the '
                                    f'TaskSequence. Please check the configuration '
                                    f'under the key "actions"!')


# -*- coding: utf-8 -*-

# Import libraries
from wasabi import Printer, TracebackPrinter
from toloka.client.task import Task
from toloka.client.pool import Pool
from toloka.metrics import MetricCollector
from typing import Union, List
import yaml
import pandas as pd
import time
import toloka.client as toloka
import toloka.metrics as metrics
import traceback
import json


# Set up Printer and TracebackPrinter
msg = Printer(pretty=True, timestamp=True, hide_animation=True)
tracep = TracebackPrinter()


def create_tasks(input_obj,
                 input_data: pd.DataFrame) -> list:
    """
    This function creates Toloka Task objects from input data based on the JSON configuration.

    Parameters:
        input_obj: A subclass of CrowdsourcingTask.
        input_data: a pandas DataFrame that contains the input data. The DataFrame should contain
                    headers that match those defined in the configuration.

    Returns:
         A list of Toloka Task objects.
    """

    # Print status message
    msg.info(f'Creating and adding tasks to pool with ID {input_obj.pool.id}')

    # Fetch input variable names from the configuration. Create a dictionary with matching
    # key and value pairs, which is updated when creating the toloka.Task objects below.
    input_values = {n: n for n in list(input_obj.conf['data']['input'].keys())}

    assert set(input_values.keys()) == set(input_data.columns), raise_error(f"Input data column names "
                                                                            f"do not match input configuration "
                                                                            f"for the pool {input_obj.name}!")

    # Create a list of Toloka Task objects by looping over the input DataFrame. Use the
    # dictionary of input variable names 'input_values' to retrieve the correct columns
    # from the DataFrame.
    tasks = [toloka.Task(pool_id=input_obj.pool.id,
                         input_values={k: row[v] for k, v in input_values.items()},
                         unavailable_for=input_obj.blocklist)
             for _, row in input_data.iterrows()]

    return tasks


def create_exam_tasks(input_obj) -> list:
    """
    This function creates Toloka Task objects with known solutions from input data based on the
    JSON configuration.

    Parameters:
        input_obj: A subclass of CrowdsourcingTask.

    Returns:
         A list of Toloka Task objects with known solutions.
    """

    # Print status message
    msg.info(f'Creating and adding exam tasks to pool with ID {input_obj.pool.id}')

    # Load exam tasks from the path defined in the JSON configuration
    exam_data = load_data(input_obj.conf['data']['file'], input_obj.conf['data']['input'])

    # Fetch input variable names from the configuration. Create a dictionary with matching
    # key and value pairs, which is updated when creating the toloka.Task objects below.
    input_values = {n: n for n in list(input_obj.conf['data']['input'].keys())}
    output_values = {n: n for n in list(input_obj.conf['data']['output'].keys())}

    # Populate the pool with exam tasks that have known answers
    tasks = [toloka.Task(pool_id=input_obj.pool.id,
                         input_values={k: row[v] for k, v in input_values.items()},
                         known_solutions=[toloka.task.BaseTask.KnownSolution(
                             output_values={k: str(row[v]) for k, v in
                                            output_values.items()})],
                         infinite_overlap=True,
                         unavailable_for=input_obj.blocklist)
             for _, row in exam_data.iterrows()]

    return tasks


def add_tasks(input_obj,
              tasks: List[Task]) -> None:
    """
    This function adds Toloka Task objects to a pool on Toloka. The function first verifies that
    these have not been added to the pool already.

    Parameters:
        input_obj: A subclass of CrowdsourcingTask.
        tasks: A list of Toloka Task objects.

    Returns:
         The Task objects are added to the pool.
    """

    # Get Task objects currently in the pool from Toloka
    old_tasks = [task for task in input_obj.client.get_tasks(pool_id=input_obj.pool.id)]

    # If the request returns Tasks from Toloka
    if len(old_tasks) > 0:

        # Compare the existing tasks to those created above under self.tasks
        tasks_exist = compare_tasks(old_tasks=old_tasks,
                                    new_tasks=tasks)

        if tasks_exist:

            # Print status message
            msg.warn(f'The tasks to be added already exist in the pool. Not adding '
                     f'duplicates.')

    else:

        # If no tasks exist, set the flag to False
        tasks_exist = False

    # If tasks do not exist in the pool
    if not tasks_exist:

        # Add tasks to the main pool
        add_tasks_to_pool(client=input_obj.client,
                          tasks=tasks,
                          pool=input_obj.pool,
                          kind='main')


def add_tasks_to_pool(client: toloka.TolokaClient,
                      tasks: list,
                      pool: Union[toloka.Pool, toloka.Training],
                      kind: str):
    """
    This function loads Toloka Task objects to a pool on Toloka.

    Parameters:
        client: A toloka.TolokaClient object with valid credentials.
        tasks: A Python list with Toloka Task objects.
        pool: A Toloka Pool object.
        kind: A string that indicates task type: 'main' or 'train'.

    Returns:
         Prints a status message to standard output.
    """
    # Attempt to create the tasks on Toloka
    try:

        if kind == 'main':

            # Create tasks on Toloka; use the default settings
            client.create_tasks(tasks, allow_defaults=True)

        if kind == 'train':

            # Create tasks on Toloka without default settings
            client.create_tasks(tasks, allow_defaults=False)

        # Print status message
        msg.good(f'Successfully added {len(tasks)} tasks to pool with ID {pool.id}')

    # Catch validation error
    except toloka.exceptions.ValidationApiError:

        # Print status message
        raise_error(f'Failed to create tasks on Toloka due to a validation error. Check '
                    f'the input data, configuration file and see the error message above! '
                    f'Note that this error was raised by {kind} tasks, so see the relevant '
                    f'section of the configuration file.')


def check_io(configuration: dict, expected_input: set, expected_output: set):

    # Read input and output data and create data specifications
    data_in = {k: data_spec[v] for k, v in configuration['data']['input'].items()}
    data_out = {k: data_spec[v] for k, v in configuration['data']['output'].items()}

    # Create a dictionary mapping input data types to variable names.
    input_data = {v: k for k, v in configuration['data']['input'].items()}

    # Raise error if the expected input data types have been provided
    if not set(input_data.keys()) == expected_input:

        raise_error(f'Could not find the expected input types ({", ".join(expected_input)}) for '
                    f'{configuration["name"]}. Please check the configuration under the '
                    f'key data/input.')

    # Create a dictionary mapping output data types to variable names.
    output_data = {v: k for k, v in configuration['data']['output'].items()}

    # Raise error if the expected input data types have been provided
    if not set(output_data.keys()) == expected_output:

        raise_error(f'Could not find the expected output types ({", ".join(expected_output)}) for '
                    f'{configuration["name"]}. Please check the configuration under the '
                    f'key data/output.')

    return data_in, data_out, input_data, output_data


def compare_tasks(old_tasks: list,
                  new_tasks: list) -> bool:
    """
    This function checks newly-defined Tasks against those that already exist in a given Pool.

    Parameters:
        old_tasks: A list of Toloka Task objects.
        new_tasks: A list of Toloka Task objects.

    Returns:
         True if the new and existing Tasks match, otherwise False
    """
    # Get the keys of existing tasks, flatten the list and cast into a set
    existing_keys = [list(task.input_values.keys()) for task in old_tasks]
    existing_keys = set([item for sublist in existing_keys for item in sublist])

    # Get the values of existing tasks, flatten the list and cast into a set
    existing_values = [list(task.input_values.values()) for task in old_tasks]
    existing_values = set([item for sublist in existing_values for item in sublist])

    # Do the same for the keys of new tasks ...
    new_keys = [list(task.input_values.keys()) for task in new_tasks]
    new_keys = set([item for sublist in new_keys for item in sublist])

    # ... and for values as well.
    new_values = [list(task.input_values.values()) for task in new_tasks]
    new_values = set([item for sublist in new_values for item in sublist])

    # Compare the sets of keys and values
    if new_keys == existing_keys and new_values == existing_values:

        # If match, return True
        return True

    else:

        return False


def get_results(client: toloka.TolokaClient,
                pool_id: str) -> pd.DataFrame:
    """
    This function retrieves task suites from a pool on Toloka, extracts the
    individual tasks and returns them in a pandas DataFrame.

    Parameters:
        client: A toloka.TolokaClient object with valid credentials.
        pool_id: A string that contains a valid pool identifier.

    Returns:
        A pandas DataFrame with assignments from the pool.
    """

    # Use the get_assignments() method to retrieve task suites from the current pool
    suites = [suite for suite in client.get_assignments(pool_id=pool_id)]

    # Define a placeholder for assignments retrieved from the task suites
    assignments = {}

    # Loop over each task suite
    for suite in suites:

        # Loop over tasks and solutions (answers) in each task suite
        for task, solution in zip(suite.tasks, suite.solutions):

            # Extract information from tasks and solutions
            assignments[len(assignments)] = {**task.input_values,
                                             **solution.output_values,
                                             'task_id': task.id,
                                             'suite_id': suite.id,
                                             'status': str(suite.status)}

    # Return results as a pandas DataFrame
    return pd.DataFrame.from_dict(assignments, orient='index')


def load_data(data: str, inputs: dict):
    """
    This function loads data from a TSV file and returns a pandas DataFrame.

    Parameters:
        data: A string that defines a path to a TSV file with data.

    Returns:
        A pandas DataFrame with input data.
    """

    # Print status
    msg.info(f'Loading data from {data}')

    # Make sure that a TSV-file is used, otherwise raise error
    assert data[-4:] == ".tsv", raise_error("Please use a TSV-file for the tasks!")

    # Load data from the TSV file into a pandas DataFrame
    try:

        # Read the TSV file – assume that header is provided on the first row
        df = pd.read_csv(data, sep='\t', header=0)

        # Convert JSON inputs from string to JSON
        json_inputs = [k for k, v in inputs.items() if v == "json"]

        for i in json_inputs:

            df[i] = df[i].apply(lambda x: json.loads(x))
            
        # Print message
        msg.good(f'Successfully loaded {len(df)} rows of data from {data}')

    except FileNotFoundError:

        # Print message
        raise_error(f'Could not load the file {data}!')

    # Return the DataFrame
    return df


def raise_error(message: str):
    """
    This function is used to raise error messages.

    Parameters:
        message: A string object that contains the error message.

    Returns:
        Raises an error.
    """
    # Create error message using TracebackPrinter
    error = tracep(title=message, tb=traceback.extract_stack())

    # Raise error
    raise ValueError(error)


def read_configuration(configuration: str):
    """
    This function reads the Task configuration from a YAML file.

    Parameters:
            configuration: A string object that defines the path to the configuration file.

    Returns:
         A Python dictionary that contains the configuration.
    """

    # Attempt to open the JSON file for reading
    try:

        with open(configuration) as conf:

            # Read the file contents and load JSON into a dictionary
            conf_dict = yaml.safe_load(conf)

            # Print status message
            msg.good(f'Successfully loaded configuration from {configuration}')

    except FileNotFoundError:

        raise_error(f'Could not load the file {configuration}! Check the path '
                    f'to the file!')

    # Return the dictionary
    return conf_dict


def set_filter(filters, new_filters):
    """
    This function checks the pool for existing filters and appends new ones to them.

    Parameters:
        filters: A Toloka filter object (e.g. Rating or FilterAnd) or None.
        new_filters: A Toloka filter object (e.g. Rating or FilterAnd).

    Returns:
        Updated Toloka filter objects for the pool.
    """
    # Check if the pool already contains filters
    if filters is not None:

        # Update the filters by adding new filters to the existing ones
        filters = (filters & new_filters)

    # Othewise set the new filters as filters
    else:

        # Update filters
        filters = new_filters

    # Return filters
    return filters


def status_change(pool: Pool) -> None:
    """
    This function monitors status changes in pools and print outs messages.

    Parameters:
        pool: A Toloka Pool object.

    Returns:
        Prints a status message to standard output.
    """
    if pool.is_closed:

        msg.info(f'Closed pool with ID {pool.id}')

    if pool.is_open:

        msg.info(f'Opened pool with ID {pool.id}')


def create_metrics(task_sequence) -> MetricCollector:
    """
    This function creates Toloka MetricCollector objects for tracking the progress of pools in a
    task sequence.

    Parameters:
        task_sequence: A TaskSequence object.

    Returns:
         A Toloka MetricCollector object.
    """
    # Set up a placeholder
    p_metrics = []

    # Create metrics for pools in the sequence
    for task in task_sequence.sequence:

        # Skip exam pools, because they run infinitely
        if hasattr(task, 'pool') and not task.exam:

            # Create metric for percentage
            p_metric = metrics.pool_metrics.PoolCompletedPercentage(pool_id=task.pool.id,
                                                                    percents_name=f'{task.name}-pct',
                                                                    toloka_client=task_sequence.client)

            # Append to the placeholder list
            p_metrics.append(p_metric)

        else:

            continue

    # Create a MetricCollector object that holds the metrics. The second argument defines the
    # function to be called on the metrics collected.
    if p_metrics:

        p_metrics = MetricCollector(p_metrics, process_metrics)

    if not p_metrics:

        p_metrics = None

    return p_metrics


def process_metrics(metric_dict: dict) -> None:
    """
    This function prints status messages about pool metrics.

    Parameters:
        metric_dict: A dictionary containing metrics for one or more pools.

    Returns:
         Prints out a status message.
    """

    for name, pct in metric_dict.items():

        msg.info(f'Pool {name.split("-")[0]} is now {pct[0][1]}% complete ...')

    time.sleep(15)


def create_pool_table(task_sequence: list) -> None:
    """
    This function creates a table with essential information on pools in a task sequence.
    A pool table. :-)

    Parameters:
        task_sequence: a list of CrowdsourcingTask objects and/or actions.

    Returns:
         Prints a table with information on all pools/actions to standard output.
    """

    # Set up headers and a placeholder for data
    header = ('Name', 'Input', 'Output', 'Pool ID', 'Project ID', 'Pool type')
    data = []

    # Loop over the tasks
    for task in task_sequence:

        # Check if training has been defined
        try:

            if task.training:

                # Collect input and output data from the configuration
                inputs = [f'{k} ({v})' for k, v in task.conf['training']['data']['input'].items()]
                outputs = [f'{k} ({v})' for k, v in task.conf['training']['data']['output'].items()]

                # Append data as a tuple to the list
                data.append((task.name, ', '.join(inputs), ', '.join(outputs), task.training.id,
                             task.training.project_id, 'Training'))

        except AttributeError:

            pass

        try:

            if task.pool:

                obj_type = 'Pool' if not task.exam else 'Exam'

                # Collect input and output data from the configuration
                inputs = [f'{k} ({v})' for k, v in task.conf['data']['input'].items()]
                outputs = [f'{k} ({v})' for k, v in task.conf['data']['output'].items()]

                # Append data as a tuple to the list
                data.append((task.name, ', '.join(inputs), ', '.join(outputs), task.pool.id,
                             task.project.id, obj_type))

        except AttributeError:

            # If task.pool raises an attribute error, the object is an action
            obj_type = 'Action'

            # Append data as a tuple to the list
            data.append((task.name, '--', '--', '--', '--', obj_type))

    # Print a table with inputs and outputs
    msg.table(data=data, header=header, divider=True, max_col=50)


def verify_connections(task_sequence: list) -> None:
    """
    This function verifies any connections between pools that have been configured.

    Parameters:
        task_sequence: A list of CrowdsourcingTask objects and/or actions.

    Returns:
        Raises an error if a pool has been referred to but cannot be found.
    """
    for task in task_sequence:

        # Check if actions have been defined in the configuration
        if 'actions' in task.conf.keys() and task.conf['actions'] is not None:

            # Check if the next task has been defined in the configuration
            try:

                # Fetch a list of tasks defined in the actions
                tasks = list(task.conf['actions'].values())

                # Check that the task exists in the task sequence
                for next_task in tasks:

                    # Check if multiple next tasks are configured and validate each individually
                    if type(next_task) == dict:

                        for n_task in next_task.values():

                            if type(n_task) == list:

                                for t in n_task:

                                    if t not in [task.name for task in task_sequence] and t not in ['accept', 'reject']:

                                        raise_error(f'Cannot find a task named {t} in the task sequence. '
                                                    f'Please check the name of the task under the key '
                                                    f'"actions" in the configuration file.')

                            else:

                                if n_task not in [task.name for task in task_sequence]:

                                    raise_error(f'Cannot find a task named {n_task} in the task sequence. '
                                                f'Please check the name of the task under the key '
                                                f'"actions" in the configuration file.')

                    elif next_task not in [task.name for task in task_sequence]:

                        raise_error(f'Cannot find a task named {next_task} in the task sequence. '
                                    f'Please check the name of the task under the key '
                                    f'"actions/next" in the configuration file.')

            except KeyError:

                pass
        
        # Check if source pool is configured to current action
        if 'source' in task.conf.keys() and task.conf['source'] is not None:

            if task.conf['source'] not in [task.name for task in task_sequence]:

                raise_error(f'Cannot find a task named {task.conf["source"]} in the task sequence. '
                            f'Please check the name of the task under the key '
                            f'"source" in the configuration file.')


def check_reward(time_per_suite: int, reward: Union[int, float], name: str) -> None:
    """
    Calculates a fair reward per task suite and checks if the configured reward reaches that.

    Parameters:
        time_per_suite: estimated time it takes a worker to complete one task suite
        reward: reward per assignment that has been set in the configuration file
        name: name of the current CrowdsourcingTask

    Returns:
        Raises a warning if the configured reward is too low and prompts the user to verify if they
        wish to proceed with the current configuration. If not, the pipeline is cancelled.
    """
    
    suites_per_hour = 60*60 / time_per_suite
    suggested_reward = 12 / suites_per_hour

    if reward < suggested_reward:

        msg.warn(f"The reward you have set per assignment for {name} does not result in a fair wage for the workers. "
                 f"In order for the workers to receive a salary of $12 per hour, set reward_per_assignment to at least ${suggested_reward}.\n"
                 f"Do you wish to proceed with the current configuration anyway (y/n)?")
        choice = input("")

        if choice == "n":

            msg.info("Cancelling pipeline", exits=1)


# Map JSON entries to Toloka objects for input/output
data_spec = {
    'url': toloka.project.UrlSpec(),
    'json': toloka.project.JsonSpec(),
    'str': toloka.project.StringSpec(),
    'bool': toloka.project.BooleanSpec()
}



# -*- coding: utf-8 -*-

# Import libraries
from ..functions.core_functions import *
from wasabi import Printer
from toloka.streaming.event import AssignmentEvent
import datetime
import uuid
import toloka.client as toloka
from toloka.client.pool.speed_quality_balance_config import BestConcurrentUsersByQuality, TopPercentageByQuality

# Set up Printer
msg = Printer(pretty=True, timestamp=True, hide_animation=True)


# Define the base class for Toloka tasks
class CrowdsourcingTask:
    """
    This is the base class for defining crowdsourcing tasks on Toloka. The class loads
    the task configuration from the JSON file, validates input data, and opens and closes
    pools on Toloka.
    """

    def __init__(self, configuration, client, task_spec, **kwargs):
        """
        This function initialises the CrowdsourcingTask class.

        Parameters:
            configuration: A dictionary that contains the configuration loaded from the YAML file.
            client: A TolokaClient object with valid credentials.
            task_spec: A Toloka TaskSpec object with task interface and input/output data.
        """
        # Create unique identifier for the Task and shorten
        self.task_id = str(uuid.uuid4())[:8].upper()

        # Assign configuration and client to attributes
        self.client = client
        self.conf = configuration

        # Unpack the configuration from YAML into attributes
        self.data_conf = self.conf['data']
        self.project_conf = self.conf['project']
        self.action_conf = self.conf['actions'] if 'actions' in self.conf.keys() else None
        self.pool_conf = self.conf['pool']
        self.train_conf = self.conf['training'] if 'training' in self.conf.keys() else None
        self.qual_conf = self.conf['quality_control'] if 'quality_control' in self.conf.keys() else None

        # Assign flags and default values
        self.name = self.conf['name']  # Unique name of the task
        self.project = None  # Placeholder for Toloka Project object
        self.training = None  # Placeholder for Toloka Training object
        self.pool = None  # Placeholder for Toloka Pool object
        self.prev_task = None  # Previous Task object in the task sequence
        self.input_data = None  # Placeholder for input data
        self.output_data = None  # Placeholder for output data
        self.train_data = None  # Placeholder for training data
        self.train_tasks = None  # Placeholder for training tasks
        self.tasks = None  # Placeholder for main tasks
        self.results = None  # Placeholder for results
        self.is_complete = False  # Is the Task complete or not?
        self.skill = False  # Does the Task provide or require a skill?
        self.exam = False  # Is this Task an exam?
        self.test = True if kwargs and kwargs['test'] else False  # Are we running a test?

        # See if users should be banned from the pool and check that blocklist is configured correctly
        try:

            self.blocklist = list(pd.read_csv(self.pool_conf['blocklist'], sep="\t")["user_id"]) \
                if 'blocklist' in self.pool_conf.keys() else []

        except KeyError:

            msg.warn(f"Could not find the column 'user_id' in the blocklist.", exits=1)

        # Print status message
        msg.info(f'The unique ID for this object ({self.name}) is {self.task_id}')

        # Get requester information unless we're running a test
        if not self.test:

            requester = client.get_requester()

            # Print status messages on requester ID and balance
            msg.info(f'Using Toloka with requester with ID {requester.id}')
            msg.info(f'Current balance on this account is ${requester.balance}')

        # Load project configuration
        self.load_project(client=client, task_spec=task_spec)

        # Load training configuration
        self.load_training(client=client)

        # Load pool configuration
        self.load_pool(client=client)

        # Create exam tasks if necessary
        if self.exam and 'file' in self.data_conf:

            # Create exam tasks and add them to the pool
            self.tasks = create_exam_tasks(self)

            if not self.test:

                add_tasks(self, tasks=self.tasks)

        # Create main tasks if they have been defined
        if not self.exam and 'file' in self.data_conf:

            # Load data
            data = load_data(self.data_conf['file'], self.data_conf['input'])

            # Create tasks and add them to the pool
            self.tasks = create_tasks(self, data)

            if not self.test:

                add_tasks(self, self.tasks)

    def __call__(self, in_obj, **options):

        # Check that the input object is a list of AssignmentEvent objects
        if type(in_obj) == list and all(isinstance(item, AssignmentEvent) for item in in_obj):

            # Check the status of AssignmentEvent objects. These AssignmentEvent objects correspond
            # to task suites on Toloka. Their status may be accepted, rejected, submitted, etc.
            for event in in_obj:

                # If the event type is accepted or submitted, create new tasks in current pool
                if event.event_type.value in ['ACCEPTED', 'SUBMITTED']:

                    # Create new Task objects
                    new_tasks = [Task(pool_id=self.pool.id,
                                      overlap=self.pool_conf['defaults']['default_overlap_for_new_tasks'],
                                      input_values={k: v for k, v in task.input_values.items()},
                                      unavailable_for=self.blocklist
                                      )
                                 for task, solution
                                 in zip(event.assignment.tasks,
                                        event.assignment.solutions)]

                    # If the assignments are for a verification pool, add the output values to the input of
                    # the new task and make the verification task unavailable for the performer
                    if options and 'verify' in options:

                        new_tasks = [Task(pool_id=self.pool.id,
                                          overlap=self.pool_conf['defaults']['default_overlap_for_new_tasks'],
                                          input_values={**task.input_values,
                                                        **solution.output_values,
                                                        'assignment_id': event.assignment.id},
                                          unavailable_for=[*self.blocklist, event.assignment.user_id])
                                     for task, solution in zip(event.assignment.tasks, event.assignment.solutions)]

                    # Add Tasks and open the pool
                    self.client.create_tasks(tasks=new_tasks, open_pool=True)

                    # Print status messages
                    msg.good(f'Received {len(new_tasks)} {event.event_type.value.lower()} tasks '
                             f'from pool {event.assignment.pool_id}')
                    msg.good(f'Creating {len(new_tasks)} new tasks in pool {self.pool.id}')

    def load_project(self, client, task_spec):
        """
        This function loads an existing project from Toloka or creates a new project according to
        the configuration in the YAML file under the key "project".

        Parameters:
            client: A toloka.TolokaClient object with valid credentials.
            task_spec: A toloka.project.task_spec.TaskSpec object that defines an user interface.

        Returns:
             Assigns a Toloka Project object under self.project.
        """

        # If the configuration file contains a key named 'id' for 'project', assume
        # that an existing project should be used.
        if 'id' in self.project_conf.keys():

            # Attempt to load the project configuration from Toloka
            try:

                # Use the get_project() method to retrieve the project
                self.project = client.get_project(project_id=self.project_conf['id'])

                # Print status message
                msg.good(f'Successfully loaded project {self.project.id} from Toloka')

            # Catch the error
            except toloka.exceptions.DoesNotExistApiError:

                # Raise error
                raise_error(f'Failed to load project with ID {self.project_conf["id"]} from '
                            f'Toloka. Check the project ID!')

        # Otherwise create new project
        else:

            # Print status message
            msg.info(f'Creating a new Toloka project ...')

            # Create a new Toloka Project object
            self.project = toloka.Project(**self.project_conf['setup'])

            try:

                # Load public instructions from a HTML file
                with open(self.project_conf['instructions'], mode='r', encoding='utf-8') as inst:

                    # Read the instructions and assign to the project
                    self.project.public_instructions = inst.read()

            except FileNotFoundError:

                # Raise error
                raise_error(f'Could not load task instructions '
                            f'from {self.project_conf["instructions"]}')

            # Assign a private comment to the project with the Task object identifier
            self.project.private_comment = f'Created by Task object with ID {self.task_id}'

            # Create the task interface
            self.project.task_spec = task_spec

            # Create new project on the platform if we're not testing
            if not self.test:

                self.project = client.create_project(self.project)

                # Print status message
                msg.good(f'Successfully created a new project with ID {self.project.id} on Toloka')

    def load_training(self, client):
        """
        This function loads a Training pool from Toloka or creates a new Training according to
        the configuration in the YAML file under the key "training".

        Parameters:
            client: A toloka.TolokaClient object with valid credentials.

        Returns:
            Assigns a Toloka Training object under self.training. If a new Training object is
            created, it is populated with Tasks as defined in the configuration.
        """

        # If a training configuration has been provided, create a training pool
        if self.train_conf is not None:

            # Check if an existing training pool should be used
            if 'id' in self.train_conf.keys():

                # Attempt to retrieve the pool from Toloka
                try:

                    # Retrieve the existing training pool from Toloka
                    self.training = client.get_training(training_id=self.train_conf['id'])

                    # Print status message
                    msg.good(f'Successfully loaded training pool with ID {self.training.id} '
                             f'from Toloka')

                # Catch the error
                except toloka.exceptions.DoesNotExistApiError:

                    # Raise error
                    raise_error(f'Failed to load training pool with ID {self.train_conf["id"]} '
                                f'from Toloka')

            # Otherwise create a new training pool
            else:

                # Create a new training
                self.training = toloka.Training(project_id=self.project.id,
                                                may_contain_adult_content=False,
                                                **self.train_conf['setup']
                                                )

                # Print status message
                msg.good(f'Successfully configured a new training pool')

                # Create the training pool if not testing
                if not self.test:

                    self.training = client.create_training(self.training)

                    # Print status messages
                    msg.good(f'Successfully created a new training pool')

                # Load training data
                self.train_data = load_data(self.conf['training']['data']['file'],
                                            self.conf['training']['data']['input'])

                # Get the input and output variable names
                input_values = {n: n for n in list(self.conf['training']['data']['input'].keys())}
                output_values = {n: n for n in list(self.conf['training']['data']['output'].keys())}

                # Create Task objects for training
                self.train_tasks = [toloka.Task(pool_id=self.training.id,
                                                input_values={k: row[v] for k, v in
                                                              input_values.items()},
                                                known_solutions=[toloka.task.BaseTask.KnownSolution(
                                                    output_values={k: str(row[v]) for k, v in
                                                                   output_values.items()})],
                                                message_on_unknown_solution=row['hint'],
                                                infinite_overlap=True,
                                                unavailable_for=self.blocklist)
                                    for _, row in self.train_data.iterrows()]

                # Add training tasks to the training pool
                if not self.test:

                    add_tasks_to_pool(client=self.client, tasks=self.train_tasks, pool=self.training,
                                      kind='train')

    def load_pool(self, client):
        """
        This function loads a main pool from Toloka or creates a new pool according to the
        configuration defined in the YAML file under the key "pool".

        Parameters:
            client: A toloka.TolokaClient object with valid credentials.

        Returns:
             Assigns a Toloka Pool object under self.pool.
        """

        # If the configuration file contains a key named 'id' under 'pool', assume
        # that an existing pool should be used for this task.
        if 'id' in self.pool_conf.keys():

            # Attempt to retrieve the main pool from Toloka
            try:

                # Retrieve the existing main pool from Toloka
                self.pool = client.get_pool(pool_id=self.pool_conf['id'])

                # Print status message
                msg.good(f'Successfully loaded main pool with ID {self.pool.id} from Toloka')

                # Check if the pool is an exam pool
                if 'exam' in self.pool_conf.keys():

                    # Set flag to True
                    self.exam = True

            # Catch the error
            except toloka.exceptions.DoesNotExistApiError:

                # Raise error
                raise_error(f'Failed to load pool with ID {self.pool_conf["id"]} from Toloka')

        # Otherwise proceed to create a new main pool
        else:

            # Set up the main pool
            self.pool = toloka.Pool(project_id=self.project.id,
                                    private_comment=f'Created by Task object with ID {self.task_id}',
                                    may_contain_adult_content=False,
                                    will_expire=datetime.datetime.now() + datetime.timedelta(
                                        days=365),
                                    **self.pool_conf['setup']
                                    )

            # Set default settings that are applied to any tasks uploaded to the pool
            self.pool.defaults = toloka.pool.Pool.Defaults(**self.pool_conf['defaults'])

            # Set the number of real, control and training tasks per suite
            self.pool.set_mixer_config(**self.pool_conf['mixer'])

            # If training has been defined, set up training and required quality
            if self.train_conf is not None:

                try:

                    # Link training pool and retrieve minimum performance value
                    self.pool.set_training_requirement(training_pool_id=self.training.id,
                                                       **self.pool_conf['training'])

                except KeyError:

                    msg.warn(f"Could not find the key 'training' under the main pool configuration. "
                             f"Define the key 'training' and place a key/value pair with the key "
                             f"'training_passing_skill_value' under 'training' to link the training "
                             f"and main pools. This key is used to set the criteria for passing "
                             f"the training, e.g. training_passing_skill_value: 70", exits=1)

            # Print status message
            msg.good(f'Successfully configured a new main pool')

            if 'estimated_time_per_suite' not in self.pool_conf.keys():

                msg.warn(f"Parameter 'estimated_time_per_suite' is not configured in pool settings "
                         f"for {self.name}. Consider adding this key under 'pool' to ensure that the "
                         f"'reward_per_assignment' you have set results in a fair hourly wage of "
                         f"at least $12. Do you wish to proceed with the current configuration "
                         f"anyway (y/n)?")

                choice = input("")

                if choice == "n":

                    msg.info("Cancelling pipeline")
                    exit()

            else:

                check_reward(self.pool_conf['estimated_time_per_suite'], 
                             self.pool_conf['setup']['reward_per_assignment'],
                             self.name)

            # Check if filters have been defined
            if 'filter' in self.pool_conf.keys() and self.pool_conf['filter'] is not None:

                # Print status message
                msg.info(f'Setting up filters')

                # Check if workers should be filtered based on language skills
                if 'languages' in self.pool_conf['filter'].keys():

                    # If only one language has been defined, proceed to set the filter
                    if len(self.pool_conf['filter']['languages']) == 1:

                        # Create filter
                        language = toloka.filter.Languages.in_(
                            self.pool_conf['filter']['languages'][0].upper())

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=language)

                    # If more than one languages have been defined, combine and set filters
                    if len(self.pool_conf['filter']['languages']) > 1:

                        # Create filters
                        languages = [(toloka.filter.Languages.in_(lang.upper()))
                                     for lang in self.pool_conf['filter']['languages']]

                        # Combine filters
                        languages = toloka.filter.FilterOr(languages)

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=languages)

                # Check if workers should be filtered based on client type
                if 'client_type' in self.pool_conf['filter'].keys():

                    # Check if only one client type has been defined
                    if len(self.pool_conf['filter']['client_type']) == 1:

                        # Create filter
                        client = (toloka.filter.ClientType ==
                                  self.pool_conf['filter']['client_type'][0].upper())

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=client)

                    # Check if more than one client type has been defined
                    if len(self.pool_conf['filter']['client_type']) > 1:

                        # Create filters
                        clients = [(toloka.filter.ClientType == client.upper())
                                   for client in self.pool_conf['filter']['client_type']]

                        # Combine filters
                        clients = toloka.filter.FilterOr(clients)

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=clients)

                # Check if workers should be filtered based on education
                if 'education' in self.pool_conf['filter'].keys():

                    # Check if only one education level has been defined
                    if len(self.pool_conf['filter']['education']) == 1:

                        # Create filter
                        education = (toloka.filter.Education ==
                                     self.pool_conf['filter']['education'][0].upper())

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=education)

                    # Check if more than one education level has been defined
                    if len(self.pool_conf['filter']['education']) > 1:

                        # Create filters
                        levels = [(toloka.filter.Education == edu.upper()) for edu in
                                  self.pool_conf['filter']['education']]

                        # Combine filters
                        levels = toloka.filter.FilterOr(levels)

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=levels)

                # Check if workers should be filtered based on skill
                if 'skill' in self.pool_conf['filter'].keys():

                    # Check if only one skill has been defined
                    if len(self.pool_conf['filter']['skill']) == 1:

                        # Get the skill ID and values from the dictionary key and cast to integers
                        skill_id = str(list(self.pool_conf['filter']['skill'][0].keys())[0])
                        skill_value = int(list(self.pool_conf['filter']['skill'][0].values())[0])

                        # Define filter
                        skill = (toloka.filter.Skill(skill_id) >= skill_value)

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=skill)

                    # Check if more than one skill has been defined
                    if len(self.pool_conf['filter']['skill']) > 1:

                        # Get the skill ID and values from the dictionary key and cast to integers
                        skills = [(toloka.filter.Skill(str(list(skill_dict.keys())[0])) >=
                                   int(list(skill_dict.values())[0]))
                                  for skill_dict in self.pool_conf['filter']['skill']]

                        # Add skills one by one
                        for skill in skills:
                            # Check for existing filters and set
                            self.pool.filter = set_filter(filters=self.pool.filter,
                                                          new_filters=skill)

                # Check if workers should be filtered based on gender
                if 'gender' in self.pool_conf['filter'].keys():

                    choice = self.pool_conf['filter']['gender'].upper()
                    gender = (toloka.filter.Gender == choice)

                    self.pool.filter = set_filter(filters=self.pool.filter,
                                                  new_filters=gender)

                # Check if workers should be filtered based on whether they allow
                # adult content or not
                if 'adult_allowed' in self.pool_conf['filter'].keys():

                    choice = bool(self.pool_conf['filter']['adult_allowed'])

                    adult_allowed = (toloka.filter.AdultAllowed == choice)

                    self.pool.filter = set_filter(filters=self.pool.filter,
                                                  new_filters=adult_allowed)

                # Check if workers should be filtered based on country
                if 'country' in self.pool_conf['filter'].keys():

                    # Check if only one country has been defined
                    if len(self.pool_conf['filter']['country']) == 1:

                        # Create filter
                        country = (toloka.filter.Country ==
                                   self.pool_conf['filter']['country'][0].upper())

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=country)

                    # Check if more than one country has been defined
                    if len(self.pool_conf['filter']['country']) > 1:

                        # Create filters
                        levels = [(toloka.filter.Country == c.upper()) for c in
                                  self.pool_conf['filter']['country']]

                        # Combine filters
                        levels = toloka.filter.FilterOr(levels)

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=levels)

                # Check if workers should be filtered based on city
                if 'city' in self.pool_conf['filter'].keys():

                    # Check if only one city has been defined
                    if len(self.pool_conf['filter']['city']) == 1:

                        # Create filter
                        city = (toloka.filter.City ==
                                   int(self.pool_conf['filter']['city'][0]))

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=city)

                    # Check if more than one city has been defined
                    if len(self.pool_conf['filter']['city']) > 1:

                        # Create filters
                        levels = [(toloka.filter.City == int(c)) for c in
                                  self.pool_conf['filter']['city']]

                        # Combine filters
                        levels = toloka.filter.FilterOr(levels)

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=levels)

                # Filter performers by date of birth (before or after a certain date or both)
                # Note that date of birth needs to be expressed as a UNIX timestamp

                # TODO converter from datetime to UNIX timestamp, so date can be configured as 
                # a date in the configuration file
                if 'date_of_birth' in self.pool_conf['filter'].keys():

                    if len(self.pool_conf['filter']['date_of_birth']) == 1:
                        
                        if 'before' in self.pool_conf['filter']['date_of_birth'].keys():
                            
                            before = self.pool_conf['filter']['date_of_birth']['before']
                            b_filter = (toloka.filter.DateOfBirth <= before)
                            self.pool.filter = set_filter(filters=self.pool.filter,
                                                          new_filters=b_filter)

                        if 'after' in self.pool_conf['filter']['date_of_birth'].keys():
                            
                            after = self.pool_conf['filter']['date_of_birth']['after']
                            a_filter = (toloka.filter.DateOfBirth >= after)
                            self.pool.filter = set_filter(filters=self.pool.filter,
                                                          new_filters=a_filter)

                    if len(self.pool_conf['filter']['date_of_birth']) > 1:

                        before = self.pool_conf['filter']['date_of_birth']['before']
                        after = self.pool_conf['filter']['date_of_birth']['after']

                        ba_filters = [(toloka.filter.DateOfBirth <= before), (toloka.filter.DateOfBirth >= after)]
                        levels = toloka.filter.FilterAnd(ba_filters)

                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=levels)

                # Check if workers should be filtered based on UserAgentType
                if 'user_agent_type' in self.pool_conf['filter'].keys():

                    # Check if only one agent type has been defined
                    if len(self.pool_conf['filter']['user_agent_type']) == 1:

                        # Create filter
                        agent = (toloka.filter.UserAgentType ==
                                   self.pool_conf['filter']['user_agent_type'][0].upper())

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=agent)

                    # Check if more than one user agent has been defined
                    if len(self.pool_conf['filter']['user_agent_type']) > 1:

                        # Create filters
                        levels = [(toloka.filter.UserAgentType == agent.upper()) for agent in
                                  self.pool_conf['filter']['user_agent_type']]

                        # Combine filters
                        levels = toloka.filter.FilterOr(levels)

                        # Check for existing filters and set
                        self.pool.filter = set_filter(filters=self.pool.filter,
                                                      new_filters=levels)

                # TODO missing filters

                # Print status message
                msg.good(f'Finished adding filters to the pool')

            # Check if skills have been defined
            if 'skill' in self.pool_conf.keys():

                # Print status message
                msg.info(f'Setting up skill')

                # Check if an existing skill has been requested
                if 'id' in self.pool_conf['skill'].keys():

                    # Attempt to retrieve the skill from Toloka
                    try:

                        # Retrieve the skill
                        self.skill = client.get_skill(skill_id=self.pool_conf['skill']['id'])

                        # Print status message
                        msg.good(f'Successfully loaded skill with ID {self.skill.id} from Toloka')

                    # Catch the error
                    except toloka.exceptions.DoesNotExistApiError:

                        # Raise error
                        raise_error(f'Failed to load skill with ID {self.pool_conf["skill"]["id"]} '
                                    f'from Toloka')

                # Otherwise create a new skill
                else:

                    # Create new skill and set public description
                    self.skill = client.create_skill(name=self.pool_conf['skill']['name'],
                                                     public_requester_description={
                                                         self.pool_conf['skill']['language']:
                                                             self.pool_conf['skill'][
                                                                 'description']})
                    # Print status message
                    msg.good(f'Successfully created skill with ID {self.skill.id}')

            # If general quality control rules exist, add them to the pool
            if self.qual_conf is not None:

                # Print status message
                msg.info(f'Setting up quality control rules')

                # Check if speed/quality balance settings have been configured
                if 'speed_quality_balance' in self.qual_conf and self.qual_conf['speed_quality_balance'] is not None:

                    if "top_percentage_by_quality" in self.qual_conf['speed_quality_balance']:
    
                        # Get percentage for top workers from configuration
                        user_percentage = self.qual_conf["speed_quality_balance"]["top_percentage_by_quality"]

                        speed_quality = TopPercentageByQuality(percent=user_percentage)
                        
                        msg.info(f"Speed/quality setting: only top {user_percentage}% of users will have access to pool {self.name}")

                    elif "best_concurrent_users_by_quality" in self.qual_conf['speed_quality_balance']:
                        
                        # Get number of top workers from configuration
                        user_count = self.qual_conf['speed_quality_balance']["best_concurrent_users_by_quality"]
                        
                        speed_quality = BestConcurrentUsersByQuality(count=user_count)

                        msg.info(f"Speed/quality setting: only top {user_count} users will have access to pool {self.name}")
                    
                    # Set speed/quality setting
                    self.pool.set_speed_quality_balance(speed_quality)

                # Set up quality control rule for fast responses
                if 'fast_responses' in self.qual_conf:

                    # Unpack rules into variables
                    history_size = self.qual_conf['fast_responses']['history_size']
                    count = self.qual_conf['fast_responses']['count']
                    threshold = self.qual_conf['fast_responses']['threshold']
                    duration = self.qual_conf['fast_responses']['ban_duration']
                    units = self.qual_conf['fast_responses']['ban_units'].upper()

                    # Add quality control rule to the pool
                    self.pool.quality_control.add_action(
                        collector=toloka.collectors.AssignmentSubmitTime(history_size=history_size,
                                                                         fast_submit_threshold_seconds=threshold),
                        conditions=[toloka.conditions.FastSubmittedCount > count],
                        action=toloka.actions.RestrictionV2(
                            scope=toloka.user_restriction.UserRestriction.PROJECT,
                            duration=duration,
                            duration_unit=units,
                            private_comment='Fast responses'
                        )
                    )

                    # Print status message
                    msg.good(f'Added quality control rule: ban for {duration} {units.lower()} if '
                             f'response time is less than {threshold} seconds for {count} out '
                             f'of {history_size} tasks')

                # Set up quality control rule for skipped assignments
                if 'skipped_assignments' in self.qual_conf:

                    # Unpack rules into variables
                    count = self.qual_conf['skipped_assignments']['count']
                    duration = self.qual_conf['skipped_assignments']['ban_duration']
                    units = self.qual_conf['skipped_assignments']['ban_units'].upper()

                    # Add quality control rule to the pool
                    self.pool.quality_control.add_action(
                        collector=toloka.collectors.SkippedInRowAssignments(),
                        conditions=[toloka.conditions.SkippedInRowCount > count],
                        action=toloka.actions.RestrictionV2(
                            scope=toloka.user_restriction.UserRestriction.PROJECT,
                            duration=duration,
                            duration_unit=units,
                            private_comment='Skipped assignments'
                        )
                    )

                    # Print status message
                    msg.good(f'Added quality control rule: ban for {duration} {units.lower()} if '
                             f'user skipped {count} assignments in a row.')

                # Set up quality control rule for re-assigning work from banned users
                if 'redo_banned' in self.qual_conf:

                    if self.qual_conf['redo_banned']:

                        # Add quality control rule to the pool
                        self.pool.quality_control.add_action(
                            collector=toloka.collectors.UsersAssessment(),
                            conditions=[toloka.conditions.PoolAccessRevokedReason
                                        == toloka.conditions.PoolAccessRevokedReason.RESTRICTION],
                            action=toloka.actions.ChangeOverlap(delta=1, open_pool=True)
                        )

                # Set up captcha quality control
                if 'captcha' in self.qual_conf:

                    # Unpack rules into variables 
                    freq = self.qual_conf['captcha']['frequency'].upper()
                    success_rate = self.qual_conf['captcha']['success_rate']
                    duration = self.qual_conf['captcha']['ban_duration']
                    units = self.qual_conf['captcha']['ban_units'].upper()

                    # Set captcha frequency according to configuration
                    self.pool.set_captcha_frequency(freq)

                    # Add quality control rule to the pool
                    self.pool.quality_control.add_action(
                        collector=toloka.collectors.Captcha(),
                        conditions=[toloka.conditions.SuccessRate < success_rate],
                        action=toloka.actions.RestrictionV2(
                            scope=toloka.user_restriction.UserRestriction.PROJECT,
                            duration=duration,
                            duration_unit=units,
                            private_comment="Too many Captcha mistakes"
                        )
                    )

                    # Print status message
                    msg.good(f"Added quality control rule: ban for {duration} {units.lower()} if "
                             f"user's CAPTCHA success rate is less than {success_rate}%. "
                             f"CAPTCHA frequency is set to {freq}.")

                # Set up GoldenSet quality control (performance on control tasks)
                if 'golden_set' in self.qual_conf:

                    # How many previous answers to control tasks to consider
                    history_size = self.qual_conf['golden_set']['history_size']

                    # Ban user if they fail control tasks with a rate higher than the threshold
                    if 'ban_rules' in self.qual_conf['golden_set']:

                        # Unpack rules into variables
                        incorrect_threshold = self.qual_conf['golden_set']['ban_rules']['incorrect_threshold']
                        ban_duration = self.qual_conf['golden_set']['ban_rules']['ban_duration']
                        ban_units = self.qual_conf['golden_set']['ban_rules']['ban_units']

                        self.pool.quality_control.add_action(
                            collector=toloka.collectors.GoldenSet(history_size=history_size),
                            conditions=[toloka.conditions.GoldenSetIncorrectAnswersRate > incorrect_threshold],
                            action=toloka.actions.RestrictionV2(
                                scope=toloka.user_restriction.UserRestriction.PROJECT,
                                duration=ban_duration,
                                duration_unit=ban_units,
                                private_comment="Fails control tasks too often"
                            )
                        )

                        # Print status message
                        msg.good(f"Added quality control rule: ban for {ban_duration} {ban_units.lower()} if "
                                 f"user fails over {incorrect_threshold}% of control tasks.")
                    
                    # Automatically reject assignments from user if they often fail control tasks
                    if 'reject_rules' in self.qual_conf['golden_set']:

                        incorrect_threshold = self.qual_conf['golden_set']['reject_rules']['incorrect_threshold']

                        self.pool.quality_control.add_action(
                            collector=toloka.collectors.GoldenSet(history_size=history_size),
                            conditions=[toloka.conditions.GoldenSetIncorrectAnswersRate > incorrect_threshold],
                            action=toloka.actions.RejectAllAssignments(
                                public_comment="Failed too many control tasks"
                            )
                        )

                        # Print status message
                        msg.good(f"Added quality control rule: Reject all user's assignments if "
                                 f"user fails over {incorrect_threshold}% of control tasks.")

                    # Automatically approve assignments from user if they are successful in control tasks
                    if 'approve_rules' in self.qual_conf['golden_set']:

                        correct_threshold = self.qual_conf['golden_set']['approve_rules']['correct_threshold']

                        self.pool.quality_control.add_action(
                            collector=toloka.collectors.GoldenSet(history_size=history_size),
                            conditions=[toloka.conditions.GoldenSetCorrectAnswersRate > correct_threshold],
                            action=toloka.actions.ApproveAllAssignments()
                        )

                        # Print status message
                        msg.good(f"Added quality control rule: Approve all user's assignments if "
                                 f"user successfully completes over {correct_threshold}% of control tasks.")

                    # Set a skill for user if they are successful in control tasks
                    if 'skill_rules' in self.qual_conf['golden_set']:
                        
                        # Unpack rules into variables
                        correct_threshold = self.qual_conf['golden_set']['skill_rules']['correct_threshold']
                        skill_id = self.qual_conf['golden_set']['skill_rules']['skill_id']
                        skill_value = self.qual_conf['golden_set']['skill_rules']['skill_value']

                        self.pool.quality_control.add_action(
                            collector=toloka.collectors.GoldenSet(history_size=history_size),
                            conditions=[toloka.conditions.GoldenSetCorrectAnswersRate > correct_threshold],
                            action=toloka.actions.SetSkill(skill_id=skill_id, skill_value=skill_value)
                        )

                        # Print status message
                        msg.good(f"Added quality control rule: Grant user a skill value {skill_value} for skill "
                                 f"{skill_id} if user successfully completes over {correct_threshold}% of control tasks.")

            # Check if the pool is an exam pool
            if 'exam' in self.pool_conf.keys():

                # Set exam flag to True
                self.exam = True

                # First, check that a skill has been defined in the configuration file
                if not self.skill:

                    # Print status message and exit
                    msg.fail(f'The pool configuration contains the key "exam", but no skill has '
                             f'been defined in the pool configuration.', exits=0)

                # Next, check that a maximum number of exam performers has been set
                if 'max_performers' not in self.pool_conf['exam']:

                    # Print status message and exit
                    msg.fail(f'The pool configuration contains the key "exam", but the number of '
                             f'maximum performers for the exam has not been set. With infinite '
                             f'overlap, the pool will never close. Set the number of maximum '
                             f'performers using the key "max_performers" under the "exam" entry.',
                             exits=0)

                # Next, check the mixer configuration – exam pools must contain tasks with known
                # answers only
                if not self.pool_conf['mixer']['real_tasks_count'] == 0 \
                        and self.pool_conf['mixer']['training_tasks_count'] == 0:

                    # Print status message and exit
                    msg.fail(f'The configuration file defines an exam pool, but the count for both '
                             f'real and training tasks is greater than 0. Exam pools must contain '
                             f'tasks with known (golden) answers only.', exits=0)

                # Finally, check that an input file for exam tasks has been provided
                if 'file' not in self.data_conf:

                    # Print status message and exit
                    msg.fail(f'The data configuration does not contain an entry for "file", which '
                             f'should define a path to a TSV with exam tasks.', exits=0)

                # Otherwise, set up the quality control rule for assigning skills
                self.pool.quality_control.add_action(
                    collector=toloka.collectors.GoldenSet(
                        history_size=self.pool_conf['exam']['history_size']),
                    conditions=[toloka.conditions.TotalAnswersCount >= self.pool_conf['exam'][
                        'min_answers']],
                    action=toloka.actions.SetSkillFromOutputField(skill_id=self.skill.id,
                                                                  from_field='correct_answers_rate'))

                # Print status message
                msg.good(f'Successfully configured exam pool using skill {self.skill.id}')

            # Create pool on Toloka
            if not self.test:

                self.pool = self.client.create_pool(self.pool)

                # Print status message
                msg.good(f'Successfully created a new pool with ID {self.pool.id} on Toloka')


# -*- coding: utf-8 -*-

# Import libraries
from ..functions.core_functions import *
from wasabi import Printer
from .core_task import CrowdsourcingTask
import toloka.client as toloka
import toloka.client.project.template_builder as tb
from collections import OrderedDict


# Set up Printer
msg = Printer(pretty=True, timestamp=True, hide_animation=True)


class ImageClassification(CrowdsourcingTask):
    """
    This is a class for image classification tasks.
    """
    def __init__(self, configuration, client, **kwargs):
        """
        This function initialises the ImageClassification class, which inherits attributes
        and methods from the superclass CrowdsourcingTask.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            An ImageClassification object.
        """
        # Read the configuration from the YAML file
        configuration = read_configuration(configuration=configuration)

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration JSON.
        super().__init__(configuration, client, task_spec, **kwargs)

    def __call__(self, input_obj, **kwargs):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj, **kwargs)

        # When called, return the ImageClassification object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url'}, {'bool'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Create the task interface; start by setting up the image viewer
        img_viewer = tb.ImageViewV1(url=tb.InputData(input_data['url']),
                                    rotatable=True,
                                    ratio=[1, 1])

        # Define the prompt text above the button group
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Set up a radio group for labels
        radio_group = tb.ButtonRadioGroupFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['bool']),

            # Create radio buttons
            options=[
                tb.fields.GroupFieldOption(value=True, label='Yes'),
                tb.fields.GroupFieldOption(value=False, label='No')
            ],

            # Set up validation
            validation=tb.RequiredConditionV1(hint="You must choose one response.")
        )

        # Set task width limit
        task_width_plugin = tb.TolokaPluginV1(kind='scroll', task_width=500)

        # Add hotkey plugin
        hotkey_plugin = tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=True),
                                           key_2=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=False))

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_viewer, prompt, radio_group]),
            plugins=[task_width_plugin, hotkey_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class ImageSegmentation(CrowdsourcingTask):
    """
    This is a class for image segmentation tasks.
    """
    def __init__(self, configuration, client):
        """
        This function initialises the ImageSegmentation class, which inherits attributes
        and methods from the superclass CrowdsourcingTask.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            An ImageSegmentation object.
        """
        # Read the configuration from the YAML file
        configuration = read_configuration(configuration=configuration)

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj, **kwargs):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj, **kwargs)

        # When called, return the ImageSegmentation object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url'}, {'json'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Create the task interface; start by setting up the image segmentation interface
        img_ui = tb.ImageAnnotationFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['json']),

            # Set up the input data field
            image=tb.InputData(input_data['url']),

            # Set up the allowed shapes: note that their order will define the order in the UI
            shapes={'rectangle': True, 'polygon': True},

            # Set this element to use all available vertical space on the page. This should ensure
            # that all UI elements are visible.
            full_height=True,

            # Set up validation
            validation=tb.RequiredConditionV1(hint="Please select at least one area!"))

        # Define the text prompt below the segmentation UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Add hotkey plugin
        hotkey_plugin = tb.ImageAnnotationHotkeysPluginV1(cancel='s',
                                                          confirm='a',
                                                          polygon='e',
                                                          rectangle='w',
                                                          select='q')

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_ui, prompt]),
            plugins=[hotkey_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class AddOutlines(CrowdsourcingTask):
    """
    This is a class for tasks that add more bounding boxes to images with pre-existing labelled bounding boxes.
    """
    def __init__(self, configuration, client):
        """
        This function initialises the AddOutlines class, which inherits attributes
        and methods from the superclass CrowdsourcingTask.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            An AddOutlines object.
        """

        # Read the configuration from the YAML file
        configuration = read_configuration(configuration=configuration)

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj, **kwargs):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj, **kwargs)

        # When called, return the AddOutlines object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url', 'json'}, {'json', 'bool'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)
        
        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)

        try:
            labels = [tb.ImageAnnotationFieldV1.Label(value=value, label=label) for 
                      value, label in configuration["interface"]["labels"].items()]

        except KeyError:

            msg.warn(f"Please add the key 'labels' under the top-level key 'interface' to define "
                     f"the labels for bounding boxes. The labels should be provided as key/value "
                     f"pairs, e.g. cat: Cat. The key must correspond to the label defined for "
                     f"the bounding box in the input JSON, whereas the value is the text displayed "
                     f"in the user interface.", exits=1)

        # Create the task interface; start by setting up the image segmentation interface
        img_ui = tb.ImageAnnotationFieldV1(

            # Set up the output data field
            data=tb.OutputData(path=output_data['json'],
                               default=tb.InputData(input_data['json'])),

            # Set up the input data field
            image=tb.InputData(input_data['url']),

            # Set up the allowed shapes: note that their order will define the order in the UI
            shapes={'polygon': True, 'rectangle': True},

            # Set this element to use all available vertical space on the page. This should ensure
            # that all UI elements are visible.
            full_height=True,

            # Set up labels for the outlines
            labels=labels,

            disabled=False
        )

        # Create a checkbox for special cases
        try:
            checkbox = tb.CheckboxFieldV1(
                data=tb.OutputData(output_data['bool'], default=False),
                label=configuration['interface']['checkbox'])

        except KeyError:
            msg.warn(f"Please add the key 'checkbox' under the top-level key 'interface' to "
                     f"define a text that is displayed above the checkbox. Define the text as a "
                     f"string e.g. checkbox: There is nothing to outline.", exits=1)

        # Define the text prompt below the segmentation UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_ui, prompt, checkbox], 
            validation=tb.AnyConditionV1(conditions=[tb.SchemaConditionV1(data=tb.OutputData(output_data['json']),
                                                                          schema={'type': 'array', 'minItems': 2}),
                                                     tb.EqualsConditionV1(data=tb.OutputData(output_data['bool']), to=True)],
                                                     hint="Outline at least one target or check the box if necessary."),
            )
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class SegmentationClassification(CrowdsourcingTask):
    """
    This is a class for binary segmentation classification tasks.
    """
    def __init__(self, configuration, client):
        """
        This function initialises the SegmentationClassification class, which inherits attributes
        and methods from the superclass Task.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            A SegmentationClassification object.
        """
        # Read the configuration from the YAML file
        configuration = read_configuration(configuration=configuration)

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj)

        # When called, return the SegmentationClassification object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url', 'json'}, {'bool'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)

        labels = [tb.ImageAnnotationFieldV1.Label(value=value, label=label) for
                  value, label in configuration["interface"]["labels"].items()] \
            if "labels" in configuration["interface"] else None

        # Create the task interface; start by setting up the image segmentation interface
        img_ui = tb.ImageAnnotationFieldV1(

            # Set up the output data field
            data=tb.InternalData(path=input_data['json'],
                                 default=tb.InputData(input_data['json'])),

            # Set up the input data field
            image=tb.InputData(path=input_data['url']),

            # Set labels
            labels=labels,

            # Set this element to use all available vertical space on the page. This should ensure
            # that all UI elements are visible.
            full_height=True,

            # Disable annotation interface
            disabled=True)

        # Define the text prompt below the segmentation UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Set up a radio group for labels
        radio_group = tb.ButtonRadioGroupFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['bool']),

            # Create radio buttons
            options=[
                tb.fields.GroupFieldOption(value=True, label='Yes'),
                tb.fields.GroupFieldOption(value=False, label='No')
            ],

            # Set up validation
            validation=tb.RequiredConditionV1(hint="You must choose one response.")
        )

        # Add hotkey plugin
        hotkey_plugin = tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=True),
                                           key_2=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=False))

        # Set task width limit
        task_width_plugin = tb.TolokaPluginV1(kind='scroll', task_width=500)

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_ui, prompt, radio_group]),
            plugins=[hotkey_plugin, task_width_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class LabelledSegmentationVerification(CrowdsourcingTask):
    """
    This is a class for binary segmentation verification tasks with labelled bounding boxes.
    """
    def __init__(self, configuration, client):
        """
        This function initialises the LabelledSegmentationVerification class, which inherits attributes
        and methods from the superclass Task.
        
        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            A LabelledSegmentationVerification object.
        """
        # Read the configuration from the YAML file
        configuration = read_configuration(configuration=configuration)

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj, verify=True)

        # When called, return the LabelledSegmentationVerification object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url', 'json', 'bool'}, {'bool'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)
        data_out['no_target'] = toloka.project.BooleanSpec()

        try:
            labels = [tb.ImageAnnotationFieldV1.Label(value=value, label=label) for 
                      value, label in configuration["interface"]["labels"].items()]

        except KeyError:

            msg.warn(f"Please add the key 'labels' under the top-level key 'interface' to define "
                     f"the labels for bounding boxes. The labels should be provided as key/value "
                     f"pairs, e.g. cat: Cat. The key must correspond to the label defined for "
                     f"the bounding box in the input JSON, whereas the value is the text displayed "
                     f"in the user interface.", exits=1)

        # Create the task interface; start by setting up the image segmentation interface
        img_ui = tb.ImageAnnotationFieldV1(

            # Set up the output data field
            data=tb.InternalData(path=input_data['json'],
                                 default=tb.InputData(input_data['json'])),

            # Set up the input data field
            image=tb.InputData(path=input_data['url']),

            # Set this element to use all available vertical space on the page. This should ensure
            # that all UI elements are visible.
            full_height=True,

            # Set labels
            labels=labels,

            # Disable annotation interface
            disabled=True)

        # Define the text prompt below the segmentation UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Create a checkbox for special cases
        try:
            checkbox = tb.CheckboxFieldV1(
                data=tb.OutputData('no_target', default=tb.InputData(input_data['bool'])),
                label=configuration['interface']['checkbox'],
                disabled=True)

        except KeyError:
            msg.warn(f"Please add the key 'checkbox' under the top-level key 'interface' to "
                     f"define a text that is displayed above the checkbox. Define the text as a "
                     f"string e.g. checkbox: There is nothing to outline.", exits=1)

        # Set up a radio group for labels
        radio_group = tb.ButtonRadioGroupFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['bool']),

            # Create radio buttons
            options=[
                tb.fields.GroupFieldOption(value=True, label='Yes'),
                tb.fields.GroupFieldOption(value=False, label='No')
            ],

            # Set up validation
            validation=tb.RequiredConditionV1(hint="You must choose one response.")
        )

        # Add hotkey plugin
        hotkey_plugin = tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=True),
                                           key_2=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=False))

        # Set task width limit
        task_width_plugin = tb.TolokaPluginV1(kind='scroll', task_width=500)

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_ui, prompt, checkbox, radio_group]),
            plugins=[hotkey_plugin, task_width_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class FixImageSegmentation(CrowdsourcingTask):
    """
    This is a class for fixing partially correct image segmentation tasks: modifying
    existing outlines and/or creating new ones.
    """
    def __init__(self, configuration, client):
        """
        This function initialises the FixImageSegmentation class, which inherits attributes
        and methods from the superclass Task.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            A FixImageSegmentation object.
        """
        # Read the configuration from the YAML file
        configuration = read_configuration(configuration=configuration)

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj, **kwargs):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj, **kwargs)

        # When called, return the FixImageSegmentation object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url', 'json'}, {'json'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)
        
        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)
        
        # Create the task interface; start by setting up the image segmentation interface
        img_ui = tb.ImageAnnotationFieldV1(

            # Set up the output data field
            data=tb.OutputData(path=output_data['json'],
                               default=tb.InputData(input_data['json'])),

            # Set up the input data field
            image=tb.InputData(input_data['url']),

            # Set up the allowed shapes: note that their order will define the order in the UI
            shapes={'rectangle': True, 'polygon': True},

            # Set this element to use all available vertical space on the page. This should ensure
            # that all UI elements are visible.
            full_height=True,

            # Set up validation
            validation=tb.RequiredConditionV1(hint="Please select at least one area!"),

            disabled=False
        )

        # Define the text prompt below the segmentation UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_ui, prompt])
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class SegmentationVerification(CrowdsourcingTask):
    """
    This is a class for binary image segmentation verification tasks.
    """
    def __init__(self, configuration, client):
        """
        This function initialises the SegmentationVerification class, which inherits attributes
        and methods from the superclass Task.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            A SegmentationVerification object.
        """
        # Read the configuration from the YAML file
        configuration = read_configuration(configuration=configuration)

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj, verify=True)

        # When called, return the SegmentationVerification object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url', 'json'}, {'bool'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)

        # Create the task interface; start by setting up the image segmentation interface
        img_ui = tb.ImageAnnotationFieldV1(

            # Set up the output data field
            data=tb.InternalData(path=input_data['json'],
                                 default=tb.InputData(input_data['json'])),

            # Set up the input data field
            image=tb.InputData(path=input_data['url']),

            # Set this element to use all available vertical space on the page. This should ensure
            # that all UI elements are visible.
            full_height=True,

            # Disable annotation interface
            disabled=True)

        # Define the text prompt below the segmentation UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Set up a radio group for labels
        radio_group = tb.ButtonRadioGroupFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['bool']),

            # Create radio buttons
            options=[
                tb.fields.GroupFieldOption(value=True, label='Yes'),
                tb.fields.GroupFieldOption(value=False, label='No')
            ],

            # Set up validation
            validation=tb.RequiredConditionV1(hint="You must choose one response.")
        )

        # Add hotkey plugin
        hotkey_plugin = tb.HotkeysPluginV1(key_1=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=True),
                                           key_2=tb.SetActionV1(data=tb.OutputData(output_data['bool']),
                                                                payload=False))

        # Set task width limit
        task_width_plugin = tb.TolokaPluginV1(kind='scroll', task_width=500)

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_ui, prompt, radio_group]),
            plugins=[hotkey_plugin, task_width_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class MulticlassVerification(CrowdsourcingTask):
    """
    This is a class for multiclass image segmentation verification tasks.
    """
    def __init__(self, configuration, client):
        """
        This function initialises the MulticlassVerification class, which inherits attributes
        and methods from the superclass Task.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            A MulticlassVerification object.
        """
        # Read the configuration from the YAML file
        configuration = OrderedDict(read_configuration(configuration=configuration))

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj, verify=True)

        # When called, return the MulticlassVerification object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'url', 'json'}, {'str'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)

        # Create the task interface; start by setting up the image interface
        img_ui = tb.ImageAnnotationFieldV1(

            # Set up the output data field
            data=tb.InternalData(path=input_data['json'],
                                 default=tb.InputData(input_data['json'])),

            # Set up the input data field
            image=tb.InputData(path=input_data['url']),

            # Set this element to use all available vertical space on the page. This should ensure
            # that all UI elements are visible.
            full_height=True,

            # Disable annotation interface
            disabled=True)

        # Define the text prompt below the image UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        options = [tb.fields.GroupFieldOption(value=value, label=label) for (value, label) 
                   in configuration['options'].items()]

        # Set up a radio group for labels
        radio_group = tb.ButtonRadioGroupFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['str']),

            # Create radio buttons
            options=options,

            # Set up validation
            validation=tb.RequiredConditionV1(hint="You must choose one response.")
        )

        # Set task width limit
        task_width_plugin = tb.TolokaPluginV1(kind='scroll', task_width=500)

        # Create hotkeys for each possible response
        hotkey_dict = {f'key_{i+1}': tb.SetActionV1(data=tb.OutputData(output_data['str']),
                                                    payload=list(configuration['options'].keys())[i]) 
                                                    for i in range(len(configuration['options']))}

        hotkey_plugin = tb.HotkeysPluginV1(**hotkey_dict)

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([img_ui, prompt, radio_group]),
            plugins=[task_width_plugin, hotkey_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class TextClassification(CrowdsourcingTask):
    """
    This is a class for text classification tasks.
    """

    def __init__(self, configuration, client):
        """
        This function initialises the TextClassification class, which inherits attributes
        and methods from the superclass CrowdsourcingTask.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            A TextClassification object.
        """
        # Read the configuration from the YAML file
        configuration = OrderedDict(read_configuration(configuration=configuration))

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj)

        # When called, return the TextClassification object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'str'}, {'str'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)

        # Create the task interface; start by setting up the text classification interface
        text_ui = tb.TextViewV1(content=tb.InputData(path=input_data['str']))

        # Define the text prompt below the classification UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        options = [tb.fields.GroupFieldOption(value=value, label=label) for (value, label) 
                   in configuration['options'].items()]

        # Set up a radio group for labels
        radio_group = tb.ButtonRadioGroupFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['str']),

            # Create radio buttons
            options=options,

            # Set up validation
            validation=tb.RequiredConditionV1(hint="You must choose one response.")
        )

        # Set task width limit
        task_width_plugin = tb.TolokaPluginV1(kind='scroll', task_width=500)

        # Create hotkeys for all possible responses
        hotkey_dict = {f'key_{i+1}': tb.SetActionV1(data=tb.OutputData(output_data['str']),
                                                    payload=list(configuration['options'].keys())[i]) 
                                                    for i in range(len(configuration['options']))}

        hotkey_plugin = tb.HotkeysPluginV1(**hotkey_dict)

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([prompt, text_ui, radio_group]),
            plugins=[task_width_plugin, hotkey_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec


class TextAnnotation(CrowdsourcingTask):
    """
    This is a class for text annotation tasks.
    """

    def __init__(self, configuration, client):
        """
        This function initialises the TextAnnotation class, which inherits attributes
        and methods from the superclass CrowdsourcingTask.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            client: A TolokaClient object with valid credentials.

        Returns:
            A TextAnnotation object.
        """
        # Read the configuration from the YAML file
        configuration = OrderedDict(read_configuration(configuration=configuration))

        # Specify task and task interface
        task_spec = self.specify_task(configuration=configuration)

        # Use the super() function to access the superclass Task and its methods and attributes.
        # This will set up the project, pool and training as specified in the configuration file.
        super().__init__(configuration, client, task_spec)

    def __call__(self, input_obj):

        # If the class is called, use the __call__() method from the superclass
        super().__call__(input_obj)

        # When called, return the TextAnnotation object
        return self

    @staticmethod
    def specify_task(configuration):
        """
        This function specifies the task interface on Toloka.

        Parameters:
            configuration: A dictionary containing the configuration defined in the YAML file.

        Returns:
             A Toloka TaskSpec object.
        """
        # Define expected input and output types for the task
        expected_i, expected_o = {'str'}, {'json'}

        # Configure Toloka data specifications and check the expected input against configuration
        data_in, data_out, input_data, output_data = check_io(configuration=configuration,
                                                              expected_input=expected_i,
                                                              expected_output=expected_o)

        # Add assignment ID to the input data
        data_in['assignment_id'] = toloka.project.StringSpec(required=False)

        # Define the text prompt above the annotation UI
        prompt = tb.TextViewV1(content=configuration['interface']['prompt'])

        # Set up annotation options
        options = [tb.fields.GroupFieldOption(value=value, label=label) for (value, label) 
                   in configuration['options'].items()]

        # Set up annotation UI
        annotation_field = tb.TextAnnotationFieldV1(

            # Set up the output data field
            data=tb.OutputData(output_data['json']),

            content=tb.InputData(input_data['str']),

            labels=options,

            # Set up validation
            validation=tb.RequiredConditionV1(hint="You must choose one response.")
        )

        # Set task width limit
        task_width_plugin = tb.TolokaPluginV1(kind='scroll', task_width=500)

        # Create hotkeys for all possible responses
        hotkey_dict = {f'key_{i+1}': tb.SetActionV1(data=tb.OutputData(output_data['json']),
                                                    payload=list(configuration['options'].keys())[i]) 
                                                    for i in range(len(configuration['options']))}

        hotkey_plugin = tb.HotkeysPluginV1(**hotkey_dict)

        # Combine the task interface elements into a view
        interface = toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1([prompt, annotation_field]),
            plugins=[task_width_plugin, hotkey_plugin]
        )

        # Create a task specification with interface and input/output data
        task_spec = toloka.project.task_spec.TaskSpec(
            input_spec=data_in,
            output_spec=data_out,
            view_spec=interface
        )

        # Return the task specification
        return task_spec



# -*- coding: utf-8 -*-

# Import libraries
from ..functions.core_functions import *
from wasabi import Printer
import contextlib
import io
import toloka.client as toloka
from toloka.streaming.event import AssignmentEvent
from toloka.client.exceptions import IncorrectActionsApiError
from typing import List, Union
import collections
import pandas as pd
import json

# Set up Printer
msg = Printer(pretty=True, timestamp=True, hide_animation=True)

f = io.StringIO()
with contextlib.redirect_stderr(f):
    from crowdkit.aggregation.classification.dawid_skene import DawidSkene
    from crowdkit.aggregation.classification.majority_vote import MajorityVote
    from crowdkit.aggregation.classification.gold_majority_vote import GoldMajorityVote
    from crowdkit.aggregation.classification.m_msr import MMSR
    from crowdkit.aggregation.classification.wawa import Wawa
    from crowdkit.aggregation.classification.zero_based_skill import ZeroBasedSkill
    from crowdkit.aggregation.classification.glad import GLAD
warn = f.getvalue()

if warn.startswith("None of PyTorch"):
    msg.warn(f"Could not find a working installation of PyTorch or TensorFlow, one of which is "
             f"needed for the crowd-kit aggregators to function. Cancelling pipeline.", exits=1)


class Verify:
    """
    This class defines an action for manually verifying crowdsourcing descriptions using other crowdsourced workers.

    To add this action to a TaskSequence, register this object with AssignmentObserver using the 'on_accepted' method.
    """
    def __init__(self, configuration, task):
        """
        This function initialises the manual verification mechanism.

        Parameters:
            configuration: A string object that defines a path to a YAML file with configuration.
            task: An object that inherits from the CrowdsourcingTask class.

        Returns:
            None
        """
        self.conf = read_configuration(configuration)
        self.name = self.conf['name']
        self.task = task
        self.client = self.task.client
        self.queue = collections.defaultdict(list)
        self.aggregator = None

    def __call__(self, events: List[AssignmentEvent]) -> None:

        # Loop over the list of incoming AssignmentEvent objects
        for event in events:

            # Zip and iterate over tasks and solutions in each event
            for task, solution in zip(event.assignment.tasks, event.assignment.solutions):

                # Retrieve the answer
                answer = solution.output_values[self.conf['data']['output']]

                # Add the answer to the queue under assignment id
                self.queue[task.input_values['assignment_id']].append(answer)

        # Set up a placeholder for processed task suites
        processed = []

        # Loop over the assignments in the queue
        for assignment_id, results in self.queue.items():

            try:

                # Accept the task suite if all assignments in the suite have been verified as correct
                if all(results) is True:

                    self.client.accept_assignment(assignment_id=assignment_id,
                                                  public_comment=self.conf['messages']['accepted'])

                    msg.good(f'Accepted assignment {assignment_id}')

                # Reject the task suite if all assignments in the suite have not been verified as correct
                if all(results) is not True:

                    self.client.reject_assignment(assignment_id=assignment_id,
                                                  public_comment=self.conf['messages']['rejected'])

                    msg.warn(f'Rejected assignment {assignment_id}')

            # Catch the error that might be raised by manually accepting/rejecting tasks in
            # the web interface
            except IncorrectActionsApiError:

                msg.warn(f'Could not {"accept" if all(results) == True else "reject"} assignment {assignment_id}')

            # Append the task suite to the list of processed suites
            processed.append(assignment_id)

        # Delete the assignment from the list of processed task suites
        for assignment_id in processed:

            processed.remove(assignment_id)


class Aggregate:
    """
    This class can be used to aggregate crowdsourced answers.

    Parameters:
        configuration: A string object that defines a path to a YAML file with configuration.
        task: Crowdsourcing Task of which results will be aggregated.
        forward: Forward object with which the aggregated tasks will be forwarded to the next pool or accepted/rejected.

    Returns:
        None
    
    """
    def __init__(self, configuration, task, forward=None):

        self.task = task
        self.conf = read_configuration(configuration)
        self.name = self.conf['name']

        self.forward = forward

        self.majority_vote = True if self.conf['method'] == 'majority_vote' else False
        self.dawid_skene = True if self.conf['method'] == 'dawid_skene' else False
        self.gold_majority_vote = True if self.conf['method'] == 'gold_majority_vote' else False
        self.mmsr = True if self.conf['method'] == 'mmsr' else False
        self.wawa = True if self.conf['method'] == 'wawa' else False
        self.zero_based_skill = True if self.conf['method'] == 'zero_based_skill' else False
        self.glad = True if self.conf['method'] == "glad" else False

        self.result = None
        self.prev_assignments = set()

        self.complete = False

    def __call__(self, pool: toloka.Pool) -> None:

        assignments = list(self.task.client.get_assignments(pool_id=pool.id, status=[toloka.Assignment.SUBMITTED, toloka.Assignment.ACCEPTED]))

        if assignments:
            a_dict = {"task": [], "inputs": [], "label": [], "worker": [], "id": []}

            input_data = list(self.task.data_conf['input'].keys())[0]
            output_data = list(self.task.data_conf['output'].keys())[0]

            for a in assignments:
                if a.id not in self.prev_assignments:
                    for i in range(len(a.tasks)):
                        a_dict['task'].append(a.tasks[i].input_values[input_data])
                        a_dict['inputs'].append(a.tasks[i].input_values)
                        a_dict['label'].append(a.solutions[i].output_values[output_data])
                        a_dict['worker'].append(a.user_id)
                        a_dict['id'].append(a.id)
                    self.prev_assignments.add(a.id)

            df = pd.DataFrame(data=a_dict)

            if self.majority_vote:

                self.result = MajorityVote().fit_predict(df)

            elif self.dawid_skene:

                self.result = DawidSkene().fit_predict(df)

            elif self.gold_majority_vote:

                raise NotImplementedError

            elif self.mmsr:

                self.result = MMSR().fit_predict(df)

            elif self.wawa:

                self.result = Wawa().fit_predict(df)

            elif self.zero_based_skill:

                self.result = ZeroBasedSkill().fit_predict(df)

            elif self.glad:

                self.result = GLAD().fit_predict(df)

            assert self.result is not None, raise_error("Aggregation did not produce a result!")

            forward_data = [{"id": df.loc[df["task"] == task, "id"].iloc[0], 
                             "input_data": df.loc[df["task"] == task, "inputs"].iloc[0], 
                             "label": self.result[task]} 
                            for task in self.result.index]

            msg.good(f"Finished aggregating {len(forward_data)} submitted tasks from {self.task.name}")
            self.complete = True

            if self.forward:
                self.forward(forward_data)


class Forward:
    """
    This class defines an action for forwarding completed tasks to specific pools based on values.

    For example, if a task receives the value True, it can be forwarded to Pool 1, whereas tasks with value False
    will be forwarded to Pool 2.

    Parameters:
        configuration: A string object that defines a path to a YAML file with configuration.
        client: Toloka client object.
        targets: Pools where tasks will be forwarded.

    Returns:
        None
    """
    def __init__(self, configuration, client, targets=None):

        self.conf = read_configuration(configuration)
        self.name = self.conf['name']
        self.client = client
        self.reject = []
        self.accept = []
        self.dont_forward = []

        # Possible outputs for the task (e.g. true and false) and their forward pools
        self.outputs = self.conf['actions']['on_result']

        # Check if some outputs should be accepted or rejected (these are not forwarded like other tasks,
        # but accepted or rejected based on the output) and remove these from outputs
        self.reject.extend([k for k, v in self.outputs.items() if v == 'reject'])
        [self.outputs.pop(k) for k in self.reject]

        self.accept.extend([k for k, v in self.outputs.items() if v == 'accept'])
        [self.outputs.pop(k) for k in self.accept]

        # If no forward pools are configured for some outputs, they will not be forwarded
        self.dont_forward.extend([k for k, v in self.outputs.items() if v == None])
        [self.outputs.pop(k) for k in self.dont_forward]

        # Deal with outputs that have several actions (both accepting/rejecting and forwarding)
        multi_action = {k: v for (k, v) in self.outputs.items() if type(v) == list}
        self.reject.extend([k for k, v in multi_action.items() if 'reject' in v])
        self.accept.extend([k for k, v in multi_action.items() if 'accept' in v])
        multi_action = {k: [i for i in v if i not in ["accept", "reject"]][0] for (k, v) in multi_action.items()}

        self.outputs = {**self.outputs, **multi_action}

        # Create mapping for output and the configured CrowdsourcingTask object of the forward pool
        self.name_mapping = {pool.name: pool for pool in targets}
        self.forward_pools = {output: self.name_mapping[name] for (output, name) in self.outputs.items()}

        # Initialize dictionary of key-list pairs. Keys are possible outputs for the task
        # and the lists are tasks to be forwarded.
        self.tasks_to_forward = collections.defaultdict(list)

    def __call__(self, events: Union[List[AssignmentEvent], List[dict]]) -> None:

        # Process tasks that come from an observer call
        if all(isinstance(x, AssignmentEvent) for x in events):

            # Loop over the list of incoming AssignmentEvent objects
            for event in events:

                for i in range(len(event.assignment.tasks)):

                    solution = event.assignment.solutions[i].output_values[self.conf['data']['output']]

                    # If performer verified the task as incorrect, reject the original assignment
                    # and, if configured in source pool under "on_reject", re-add the task to the pool
                    if solution in self.reject:

                        self.client.reject_assignment(assignment_id=event.assignment.tasks[i].input_values['assignment_id'],
                                                      public_comment="Assignment was verified as incorrect by another user.")
                        msg.warn(f'Rejected assignment {event.assignment.tasks[i].input_values["assignment_id"]}')

                    # If performer verified the task as correct, accept original assignment and don't forward task
                    if solution in self.accept:

                        self.client.accept_assignment(assignment_id=event.assignment.tasks[i].input_values['assignment_id'],
                                                    public_comment="Assignment was verified correct by another user.")
                        msg.good(f'Accepted assignment {event.assignment.tasks[i].input_values["assignment_id"]}')

                    # If no forward pool was configured, submit task without forwarding/accepting/rejecting
                    if solution in self.dont_forward:

                        msg.good(f'Received a submitted assignment with output "{solution}"')

                    # Else, forward task according to configuration
                    if solution in self.forward_pools:

                        try:
                            task = toloka.Task(
                                pool_id = self.forward_pools[solution].pool.id,
                                input_values=event.assignment.tasks[i].input_values,
                                unavailable_for=self.forward_pools[solution].blocklist
                            )
                            self.tasks_to_forward[solution].append(task)

                        # Catch errors
                        except toloka.exceptions.ValidationApiError:

                            # Raise error
                            raise_error(f'Failed to forward assignment {event.assignment.tasks[i].input_values["assignment_id"]}')

                        # If object has no attribute 'pool', it is an action, and should be called to activate
                        except AttributeError:

                            action = self.forward_pools[solution]
                            action(event)

            tasks_list = [task for l in self.tasks_to_forward.values() for task in l]

            if tasks_list:
                
                # Add tasks to defined pools
                self.client.create_tasks(tasks_list, allow_defaults=True, open_pool=True)

                # Print status if any tasks were forwarded on this call
                msg.good(f"Successfully forwarded {len(tasks_list)} {'tasks' if len(tasks_list) > 1 else 'task'}")

            # Tasks currently in lists have been forwarded, so reset lists
            self.tasks_to_forward = collections.defaultdict(list)
            tasks_list = []

        # Process tasks that come from aggregation
        if all(isinstance(x, dict) for x in events):

            # Loop over the list of incoming AssignmentEvent objects
            for event in events:

                solution = event['label']

                # If performer verified the task as incorrect, reject the original assignment
                # and, if configured in source pool under "on_reject", re-add the task to the pool
                if solution in self.reject:

                    self.client.reject_assignment(assignment_id=event["input_data"]["assignment_id"],
                                                  public_comment="Assignment was verified incorrect by another user.")
                    msg.warn(f'Rejected assignment {event["input_data"]["assignment_id"]}')

                # If performer verified the task as correct, accept original assignment and don't forward task
                if solution in self.accept:

                    self.client.accept_assignment(assignment_id=event["input_data"]["assignment_id"],
                                                  public_comment="Assignment was verified correct by another user.")
                    msg.good(f'Accepted assignment {event["input_data"]["assignment_id"]}')

                # If no forward pool was configured, submit task without forwarding/accepting/rejecting
                if solution in self.dont_forward:

                    msg.good(f'Received a submitted assignment with output "{solution}"')

                # Else, forward task according to configuration
                if solution in self.forward_pools:

                    try:
                        task = toloka.Task(
                            pool_id = self.forward_pools[solution].pool.id,
                            input_values=event['input_data'],
                            unavailable_for=self.forward_pools[solution].blocklist
                        )
                        self.tasks_to_forward[solution].append(task)

                    # Catch errors
                    except toloka.exceptions.ValidationApiError:

                        # Raise error
                        raise_error(f'Failed to forward aggregated task')

                    # If object has no attribute 'pool', it is an action, and should be called to activate
                    except AttributeError:

                        action = self.forward_pools[solution]
                        action(event)
                
            tasks_list = [task for l in self.tasks_to_forward.values() for task in l]

            if tasks_list:
                
                # Add tasks to defined pools
                self.client.create_tasks(tasks_list, allow_defaults=True, open_pool=True)

                # Print status if any tasks were forwarded on this call
                msg.good(f"Successfully forwarded {len(tasks_list)} {'tasks' if len(tasks_list) > 1 else 'task'}")

            # Tasks currently in lists have been forwarded, so reset lists
            self.tasks_to_forward = collections.defaultdict(list)
            tasks_list = []


class SeparateBBoxes:
    """
    This class defines an action for separating an image with several bounding boxes to several tasks with one
    bounding box per image.

    Parameters:
        target: CrowdsourcingTask to which the new tasks will be created
        configuration: A string object that defines a path to a YAML file with configuration.
        add_label: A string object of a label that will be given to the created bounding boxes. Optional parameter.

    Returns:
        None
    """

    def __init__(self, target, configuration, add_label=False):
        self.target = target
        self.client = self.target.client
        self.conf = read_configuration(configuration)
        self.name = self.conf["name"]
        self.add_label = add_label

        if "input_file" in self.conf["data"]:
            self.input_file = self.conf["data"]["input_file"]

    def __call__(self, event: Union[AssignmentEvent, dict, List[AssignmentEvent]]=None) -> None:

        # If the object is registered with an observer, use data from AssignmentEvents or the incoming dict to create new tasks
        if event:

            if type(event) == AssignmentEvent:

                msg.info(f"Creating new tasks in pool {self.target.name} with action {self.name}")

                if self.add_label:

                    for i in range(len(event.assignment.tasks)):

                        new_tasks = [toloka.Task(pool_id=self.target.pool.id,
                                                input_values={"image": event.assignment.tasks[i].input_values["image"], "outlines": [bbox]},
                                                unavailable_for=self.target.blocklist) 
                                        for bbox in [dict(x, **{'label': self.add_label}) for x in event.assignment.solutions[i].output_values["outlines"]] ]

                        self.client.create_tasks(new_tasks, allow_defaults=True, open_pool=True)

                else:

                    for i in range(len(event.assignment.tasks)):

                        new_tasks = [toloka.Task(pool_id=self.target.pool.id,
                                                input_values={"image": event.assignment.tasks[i].input_values["image"], "outlines": [bbox]},
                                                unavailable_for=self.target.blocklist) 
                                        for bbox in event.assignment.solutions[i].output_values["outlines"] ]

                        self.client.create_tasks(new_tasks, allow_defaults=True, open_pool=True)

            elif type(event) == List[AssignmentEvent]:

                msg.info(f"Creating new tasks in pool {self.target.name} with action {self.name}")

                if self.add_label:

                    for e in event:

                        for i in range(len(e.assignment.tasks)):

                            new_tasks = [toloka.Task(pool_id=self.target.pool.id,
                                                    input_values={"image": e.assignment.tasks[i].input_values["image"], "outlines": [bbox]},
                                                    unavailable_for=self.target.blocklist) 
                                            for bbox in [dict(x, **{'label': self.add_label}) for x in e.assignment.solutions[i].output_values["outlines"]] ]

                            self.client.create_tasks(new_tasks, allow_defaults=True, open_pool=True)

                else:

                    for e in event:

                        for i in range(len(e.assignment.tasks)):

                            new_tasks = [toloka.Task(pool_id=self.target.pool.id,
                                                    input_values={"image": e.assignment.tasks[i].input_values["image"], "outlines": [bbox]},
                                                    unavailable_for=self.target.blocklist) 
                                            for bbox in e.assignment.solutions[i].output_values["outlines"] ]

                            self.client.create_tasks(new_tasks, allow_defaults=True, open_pool=True)

            # If input comes from a forward action, it is in the form of a dictionary:
            elif type(event) == dict:

                msg.info(f"Creating new tasks in pool {self.target.name} with action {self.name}")

                if self.add_label:

                    new_tasks = [toloka.Task(pool_id=self.target.pool.id,
                                            input_values={"image": event["input_data"]["image"], "outlines": [bbox]},
                                            unavailable_for=self.target.blocklist) 
                                        for bbox in [dict(x, **{'label': self.add_label}) for x in event['input_data']['outlines']] ]

                    self.client.create_tasks(new_tasks, allow_defaults=True, open_pool=True)

                else:

                    new_tasks = [toloka.Task(pool_id=self.target.pool.id,
                                            input_values={"image": event["input_data"]["image"], "outlines": [bbox]},
                                            unavailable_for=self.target.blocklist) 
                                        for bbox in event['input_data']['outlines'] ]

                    self.client.create_tasks(new_tasks, allow_defaults=True, open_pool=True)

        # If the object is called without AssignmentEvents, the action starts the pipeline and
        # input data should be read from a file 
        else:
        
            input_df = pd.read_csv(self.input_file, sep="\t", index_col=False)

            msg.info(f"Creating new tasks in pool {self.target.name} with action {self.name}")

            if self.add_label:

                input_df["outlines"] = input_df["outlines"].apply(lambda x: json.loads(x))
                input_df["outlines"] = input_df["outlines"].apply(lambda x: [dict(y, **{'label': self.add_label}) for y in x])

            for i, task in input_df.iterrows():

                new_tasks = [toloka.Task(pool_id=self.target.pool.id,
                                        input_values={"image": task["image"], "outlines": [bbox]},
                                        unavailable_for=self.target.blocklist)
                            for bbox in task["outlines"]
                ]
                
                self.client.create_tasks(new_tasks, allow_defaults=True, open_pool=True)



# -*- coding: utf-8 -*-

# Import libraries
from wasabi import Printer
from toloka.streaming.observer import BaseObserver
from toloka.client.analytics_request import UniqueWorkersCountPoolAnalytics, ActiveWorkersByFilterCountPoolAnalytics, \
    SubmittedAssignmentsCountPoolAnalytics
from toloka.client.operations import Operation
from toloka.client.exceptions import DoesNotExistApiError


# Set up Printer
msg = Printer(pretty=True, timestamp=True, hide_animation=True)


class AnalyticsObserver(BaseObserver):

    def __init__(self, client, pool, **options) -> None:
        super().__init__()
        self.name = pool.id
        self.client = client
        self.operation = None
        self.pool = pool
        self.limit = options['max_performers'] if options and 'max_performers' in options else None
        self.limit_reached = False
        self.prev_result = None

    async def __call__(self):

        pass

    async def should_resume(self) -> bool:

        if self.operation is not None:

            # Fetch operation from Toloka
            operation = await self.client.get_operation(self.operation.id)

            # Check status
            if operation.status == Operation.Status.SUCCESS:

                for response in operation.details['value']:

                    # Check if pool should be closed after certain number of workers have
                    # submitted to the pool (this is used to limit the number of workers
                    # for exam pools with infinite overlap)
                    if self.limit and not self.limit_reached:

                        if response['request']['name'] == 'unique_workers_count' \
                                and response['result'] >= self.limit:

                            msg.warn(f'Maximum number of performers ({self.limit}) reached for '
                                     f'pool {self.pool.id}; closing pool ...')

                            self.limit_reached = True

                            await self.client.close_pool_async(pool_id=self.pool.id)

                            msg.good(f'Successfully closed pool {self.pool.id}')

                            # Manually close training pool(s) when exam pool is closed
                            try:

                                for training in await self.client.get_trainings(project_id=self.pool.project_id):
                                    await self.client.close_training_async(training_id=training.id)
                                    msg.good(f"Successfully closed training pool {training.id}")

                            except StopIteration:
                                
                                # If no training is found, continue as normal
                                pass

                            except DoesNotExistApiError:

                                # If a training is found, but it cannot be closed, raise warning
                                msg.warn(f"Attempted to close a training for pool {self.pool.id} that does not exist")

                    if response['request']['name'] == 'unique_workers_count' and not self.limit_reached:

                        # Check if number of submissions has changed: only print update 
                        # if there has been a change.
                        if response['result'] != self.prev_result:
                            
                            msg.info(f'{response["result"]} workers submitted to pool {self.pool.id}')
                            self.prev_result = response["result"]

            # If the operation is completed, reset the operation
            if operation.status.value in ['SUCCESS', 'FAIL']:

                self.operation = None

        if self.operation is None:

            # Create a new operation for analytics request
            self.operation = await self.create_operation()

        # Return False to prevent running forever
        return False

    async def create_operation(self):
        """
        This function creates an operation for requesting pool analytics from Toloka.

        Returns:
            An Operation object.
        """
        # Define analytics to be requested
        stat_requests = [
            UniqueWorkersCountPoolAnalytics(subject_id=self.pool.id),
            ActiveWorkersByFilterCountPoolAnalytics(subject_id=self.pool.id, interval_hours=1),
            SubmittedAssignmentsCountPoolAnalytics(subject_id=self.pool.id)
        ]

        # Get the analytics and return
        operation = await self.client.get_analytics(stat_requests)

        return operation
